# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates and trains one class classifiers to use in the SPADE architecture."""

# Using typing instead of collections due to Vertex training containers
# not supporting them.

from typing import MutableMapping, Sequence, Optional

from absl import logging
import numpy as np
from sklearn import mixture
import tensorflow as tf


# TODO(b/247116870): Create abstract class for templating out future OCC models.
class GmmEnsemble:
  """Class for creating and training a Gaussian mixture model ensemble.

  Initializes attributes for creating single GMMs used in the ensemble, as well
  as methods for fitting models and scoring new data.

  Attributes:
    n_components: Number of mixture components.
    covariance_type: Covariance matrix to create, can be one of full, diag,
      spherical, or tied.
    init_params: Method used to initialize the input matrix. For kmeans, the k
      will be equal to the number of mixture components (n_components).
    max_iter: Maximum number of loops to run in the EM algorithm.
    ensemble_count: Number of Gaussian mixture models to create and run on the
      entire training dataset. This will be used to score samples and classify
      them as anomalies. Note that the more models in the ensemble, the harder
      it is to gain consensus.
    positive_threshold: Float between [0, 100] used as the percentile for the
      one class classifier ensemble to label a point as positive. The closer to
      0 this value is set, the less positive data will be labeled. However, we
      expect an increase in precision when lowering the value, and an increase
      in recall when raising it. Equavalent to saying the given data point needs
      to be located in the top X percentile in order to be considered anomalous.
    negative_threshold: Float between [0, 100] used as the percentile for the
      one class classifier ensemble to label a point as negative. The higher
      this value is set, the less negative data will be labeled. A value in the
      range of 60-95 is a good starting point. We expect an increase in
      precision when raising this value, and an increase in recall when lowering
      it. Equavalent to saying the given data point needs to be X percentile or
      greater in order to be considered anomalous.
    verbose: Boolean denoting whether to send model performance and
      pseudo-labeling metrics to the GCP console.
    ensemble: A trained ensemble of one class classifiers.
  """

  # TODO(b/247116870): Create dataclass when another OCC is added.
  def __init__(
      self,
      n_components: int = 1,
      covariance_type: str = 'full',
      init_params: str = 'kmeans',
      max_iter: int = 100,
      ensemble_count: int = 5,
      positive_threshold: float = 1.0,
      negative_threshold: float = 95.0,
      verbose: bool = False,
  ) -> None:
    self.n_components = n_components
    self.covariance_type = covariance_type
    self.init_params = init_params
    self.max_iter = max_iter
    self.ensemble_count = ensemble_count
    self.positive_threshold = positive_threshold
    self.negative_threshold = negative_threshold
    self.verbose = verbose

    self.ensemble = []

    self._warm_start = False

  def _get_model(self) -> mixture.GaussianMixture:
    """Instantiates a Gaussian mixture model.

    Returns:
      Gaussian mixture model with class attributes.
    """
    return mixture.GaussianMixture(
        n_components=self.n_components,
        covariance_type=self.covariance_type,
        init_params=self.init_params,
        warm_start=self._warm_start,
        max_iter=self.max_iter,
    )

  def fit(
      self, train_x: tf.data.Dataset, batches_per_occ: int
  ) -> Sequence[mixture.GaussianMixture]:
    """Creates and fits and ensemble of one class classifiers.

    This model should only be called using the negative and unlabeled data.

    Args:
      train_x: Tensorflow dataset to use as input in creating the OCCs. This
        dataset will be used in training all models in the ensemble.
      batches_per_occ: The number of training batches to use in fitting each
        OCC.

    Returns:
      A list containing all trained models in the ensemble.
    """
    # If there is just one batch for a single OCC, we can leave
    # warm_start=False, else we need it enabled.
    if batches_per_occ > 1:
      self._warm_start = True

    dataset_iterator = train_x.as_numpy_iterator()

    for _ in range(self.ensemble_count):
      model = self._get_model()

      for _ in range(batches_per_occ):
        features, labels = dataset_iterator.next()
        del labels  # Not needed for this task.
        model.fit(features)

      self.ensemble.append(model)

    # Reset warm start in case fit is called multiple times with different
    # parameters on the same GmmEnsemble object.
    self._warm_start = False

    return self.ensemble

  def _score_unlabeled_data(
      self,
      unlabeled_features: np.ndarray,
  ) -> MutableMapping[str, np.ndarray]:
    """Labels data as a positive (anomalous) or negative (normal).

    Given an unlabled dataset and percentiles, determines the indices of the
    dataset that are believed to be positive or normal samples. The dataset will
    then be updated and uploaded to BigQuery where a supervised model will be
    trained and deployed. Default percentiles should be changed depending on the
    data that is passed in, and can have a significant effect on the
    performance. Also note that every model in the ensemble needs to agree in
    order for a data point to be classified as anomalous or normal.

    Args:
      unlabeled_features: Numpy array consisting of unlabeled data.

    Returns:
      A mutable mapping composed of positive indices and negative indices for
      the unlabled dataset.
    """
    # TODO(b/247116870): Update this method to take a batch size so that we can
    # perform inference on large datasets. Ideally, everything will be done
    # using indices, and data is updated in BigQuery with the new label.

    # Initialize arrays that we will use to keep the score for each row vector.
    model_scores_pos = np.zeros(len(unlabeled_features))
    model_scores_neg = np.zeros(len(unlabeled_features))

    for model in iter(self.ensemble):
      unlabeled_scores = model.score_samples(unlabeled_features)

      thresh_pos = np.percentile(unlabeled_scores, self.positive_threshold)
      thresh_neg = np.percentile(unlabeled_scores, self.negative_threshold)

      binary_scores_pos = (unlabeled_scores < thresh_pos)
      binary_scores_neg = (unlabeled_scores > thresh_neg)

      model_scores_pos += binary_scores_pos
      model_scores_neg += binary_scores_neg

    positive_indices = np.where(model_scores_pos == self.ensemble_count)[0]
    negative_indices = np.where(model_scores_neg == self.ensemble_count)[0]

    return {
        'positive_indices': positive_indices,
        'negative_indices': negative_indices
    }

  def pseudo_label(self,
                   features: np.ndarray,
                   labels: np.ndarray,
                   positive_data_value: int,
                   negative_data_value: Optional[int],
                   unlabeled_data_value: int,
                   alpha: float = 1.0,
                   verbose: Optional[bool] = False) -> Sequence[np.ndarray]:
    """Labels unlabeled data using the trained ensemble of OCCs.

    Args:
      features: A numpy array of features, already pre-processed using a min-max
        scaler.
      labels: A numpy array of labels. Strings or integers can be used for
        denoting the positive, negative, or unlabeled features.
      positive_data_value: The value used in the label column to denote
        positive data - data points that are anomalous.
      negative_data_value: The value used in the label column to denote
        negative data - data points that are not anomalous.
      unlabeled_data_value: The value used in the label column to denote
        unlabeled data.
      alpha: This value is used to adjust the influence of the pseudo labeled
        data in training the supervised model.
      verbose: Chooses the amount of logging info to display. This can be useful
        when debugging model performance.

    Returns:
      A sequence including updated features (features for which we now have
      labels for), updated labels (includes pseudo labeled positive and negative
      values, as well as ground truth), and the weights (correct alpha values)
      for the new pseudo labeled data points. Labels are in the format of 1 for
      positive and 0 for negative.
    """
    original_positive_idx = np.where(labels == positive_data_value)[0]
    original_negative_idx = np.where(labels == negative_data_value)[0]
    original_unlabeled_idx = np.where(labels == unlabeled_data_value)[0]

    updated_indices = self._score_unlabeled_data(
        unlabeled_features=features[original_unlabeled_idx, :])

    new_positive_indices = updated_indices['positive_indices']
    new_negative_indices = updated_indices['negative_indices']

    new_positive_features = features[
        original_unlabeled_idx[new_positive_indices]]
    new_negative_features = features[
        original_unlabeled_idx[new_negative_indices]]
    positive_features = features[original_positive_idx]
    negative_features = features[original_negative_idx]

    # Build the new feature, label, and weight set to send to a supervised
    # model.
    new_features = np.concatenate([
        new_positive_features,
        new_negative_features,
        positive_features,
        negative_features,
    ],
                                  axis=0)
    new_labels = np.concatenate([
        np.ones(len(new_positive_indices)),
        np.zeros(len(new_negative_indices)),
        np.ones(len(original_positive_idx)),
        np.zeros(len(original_negative_idx))
    ],
                                axis=0)
    weights = np.concatenate([
        np.repeat(alpha, len(new_positive_indices)),
        np.repeat(alpha, len(new_negative_indices)),
        np.ones([len(original_positive_idx)]),
        np.ones([len(original_negative_idx)])
    ],
                             axis=0)

    if verbose:
      logging.info('Number of new positive labels: %s',
                   len(new_positive_indices))
      logging.info('Number of new negative labels: %s',
                   len(new_negative_indices))

    return new_features, new_labels, weights



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for creating a tf.data.Dataset from a BigQuery table."""

import collections
import dataclasses
import functools
import tempfile
import typing
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from google.cloud import bigquery
from google.cloud import bigquery_storage
import numpy as np
import pandas as pd
import tensorflow as tf

# pylint: disable=g-statement-before-imports,g-import-not-at-top
if typing.TYPE_CHECKING:
  from spade_anomaly_detection.data_utils import bq_utils
  from spade_anomaly_detection.data_utils import feature_metadata
else:
  try:
    from spade_anomaly_detection.data_utils import bq_utils
    from spade_anomaly_detection.data_utils import feature_metadata
  except ImportError:
    import bq_utils
    import feature_metadata
# pylint: enable=g-statement-before-imports,g-import-not-at-top


# TODO(b/333154677): Investigate using BigQueryReadClient with one generator per
#  stream.

NO_CACHE_LOCATION_NAME = 'none'

VALUES_KEY = 'values'
WAS_NULL_KEY = 'was_null'

# The value used to replace NULL strings.
NULL_STRING_PLACEHOLDER = '__NULL_PLACEHOLDER__'
# The value used to replace NULL floats.
NULL_FLOAT_PLACEHOLDER = 0.0
# The value used to replace NULL integers.
NULL_INT_PLACEHOLDER = 0
# The value used to replace NULL booleans.
NULL_BOOL_PLACEHOLDER = False

TensorAndMaskType = Union[
    Mapping[str, tf.Tensor], Mapping[str, Mapping[str, tf.Tensor]]
]
TensorAndMaskSpecType = Union[
    Mapping[str, tf.TensorSpec], Mapping[str, Mapping[str, tf.TensorSpec]]
]
KerasInputType = Callable[..., tf.Tensor]

_TF_FLOAT_DTYPE = tf.dtypes.float32
_TF_INT_DTYPE = tf.dtypes.int32
_TF_BOOL_DTYPE = tf.dtypes.bool
_NP_FLOAT_DTYPE = np.float32
_NP_INT_DTYPE = np.int32
_NP_BOOL_DTYPE = np.bool_

# The default value for variables related to input formatting.
NESTED_FORMAT_DEFAULT = False

_USE_LEGACY_SQL = False
_WITH_MASK_DEFAULT = False


def convert_series_to_tensor_dictionary(
    input_series: pd.Series,
    value_to_replace_null: Any,
    np_dtype: np.dtype,
    tf_dtype: tf.dtypes.DType,
    with_mask: bool = _WITH_MASK_DEFAULT,
) -> Union[tf.Tensor, Mapping[str, tf.Tensor]]:
  """Converts the input series into a tensor after filling null values.

  Args:
    input_series: The series to be converted.
    value_to_replace_null: Any nulls in this feature will be replaced with this
      value.
    np_dtype: The data in the series will be converted to this data type after
      filling and nulls and prior to converting to Tensorflow.
    tf_dtype: The data type for the output tensor(s).
    with_mask: If True (the default) the data will be output as a dictionary
      with the keys 'values' and 'was_null'. Otherwise, the tensor will be
      output directly.

  Returns:
    A tensor with the null values filled. If with_mask is true this tensor
    will be included in a dictionary with the key 'values' and the dictionary
    will also contain the key 'was_null' which will be a boolean tensor that
    is true if that value was pandas NA and false otherwise.
  """
  if with_mask:
    # TODO(b/333154677): Check performance to see if using null mask is faster.
    missing_mask = input_series.isnull()
    # pandas disabled silent downcasting in version 2.2.1, suppressing
    # the warning here since we auto-inferred the dtype already.
    # Uncomment this when pandas version is updated to 2.2.1:
    # with pd.option_context('future.no_silent_downcasting', True):
    input_series = input_series.fillna(value_to_replace_null)
    return {
        VALUES_KEY: tf.convert_to_tensor(
            input_series.to_numpy(dtype=np_dtype),
            dtype=tf_dtype,
        ),
        WAS_NULL_KEY: tf.convert_to_tensor(
            missing_mask.to_numpy(dtype=bool),
            dtype=tf.dtypes.bool,
        ),
    }
  else:
    # Uncomment this when pandas version is updated to 2.2.1:
    # with pd.option_context('future.no_silent_downcasting', True):
    filled_inputs = input_series.fillna(value_to_replace_null)
    return tf.convert_to_tensor(
        filled_inputs.to_numpy(dtype=np_dtype),
        dtype=tf_dtype,
    )


# noinspection PyUnresolvedReferences
@dataclasses.dataclass
class BQFeatureConverter:
  """Converts an input series into a structure that can be used by tf.Dataset.

  Uses the input arguments to help convert the incoming data (i.e. pd.Series)
  into a dictionary of tensors (or dictionary of dictionaries) that can be
  passed into a TensorFlow dataset.

  Attributes:
    value_to_replace_null: Any nulls in this feature will be replaced with this
      value.
    np_dtype: The data in the series will be converted to this data type after
      filling and nulls and prior to converting to Tensorflow.
    tf_dtype: The data type for the output.
  """

  value_to_replace_null: Any
  np_dtype: np.dtype
  tf_dtype: tf.dtypes.DType

  def series_to_tensor(
      self,
      input_series: pd.Series,
      with_mask: bool = _WITH_MASK_DEFAULT,
  ) -> Union[tf.Tensor, Mapping[str, tf.Tensor]]:
    """Converts the input series into a tensor after filling null values.

    Args:
      input_series: The series to be converted.
      with_mask: If True (the default) the data will be output as a dictionary
        with the keys 'values' and 'was_null'. Otherwise the tensor will be
        output directly.

    Returns:
      A tensor with the null values filled. If with_mask is true this tensor
      will be included in a dictionary with the key 'values' and the dictionary
      will also contain the key 'was_null' which will be a boolean tensor that
      is true if that value was pandas NA and false otherwise.
    """
    return convert_series_to_tensor_dictionary(
        input_series,
        self.value_to_replace_null,
        self.np_dtype,
        self.tf_dtype,
        with_mask,
    )


# https://numpy.org/doc/stable/reference/arrays.dtypes.html#string-dtype-note
_STRING_CONVERSION_INFO = BQFeatureConverter(  # pytype: disable=wrong-arg-types  # numpy-scalars
    NULL_STRING_PLACEHOLDER,
    np.str_,
    tf.dtypes.string,
)
_FLOAT_CONVERSION_INFO = BQFeatureConverter(  # pytype: disable=wrong-arg-types  # numpy-scalars
    NULL_FLOAT_PLACEHOLDER,
    _NP_FLOAT_DTYPE,
    _TF_FLOAT_DTYPE,
)
_INT_CONVERSION_INFO = BQFeatureConverter(  # pytype: disable=wrong-arg-types  # numpy-scalars
    NULL_INT_PLACEHOLDER, _NP_INT_DTYPE, _TF_INT_DTYPE
)
_BOOL_CONVERSION_INFO = BQFeatureConverter(  # pytype: disable=wrong-arg-types  # numpy-scalars
    NULL_BOOL_PLACEHOLDER, _NP_BOOL_DTYPE, _TF_BOOL_DTYPE
)

# Create a mapping from the BigQuery data types to how we want to handle their
# conversion to tensors.  This uses a default type of converting the objects to
# strings if the conversion is not explicitly included in the dictionary.
_TF_INFO_FROM_BQ_DTYPE = collections.defaultdict(
    lambda: _STRING_CONVERSION_INFO,
    {
        'BIGNUMERIC': _FLOAT_CONVERSION_INFO,
        'FLOAT': _FLOAT_CONVERSION_INFO,
        'FLOAT64': _FLOAT_CONVERSION_INFO,
        'INT64': _INT_CONVERSION_INFO,
        'INTEGER': _INT_CONVERSION_INFO,
        'NUMERIC': _FLOAT_CONVERSION_INFO,
        'STRING': _STRING_CONVERSION_INFO,
        'BOOL': _BOOL_CONVERSION_INFO,
    },
)


# TODO(b/333154677): Make this cleaner and more dynamic.
def _dataframe_to_dict_of_tensors(
    df: pd.DataFrame,
    metadata_container: feature_metadata.FeatureMetadataContainer,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
) -> TensorAndMaskType:
  """Converts a dataframe to a dictionary of tensors.

  Args:
    df: The input dataframe to be converted.
    metadata_container: The metadata collection for the data in the dataframe.
    with_mask: If true a mask specifying if the data was null will also be
      returned.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.

  Returns:
    A dictionary of tensors and potentially masks of each of the features in the
    input dataframe.
  """
  output = {}
  for col_name, col_data in df.items():
    metadata = metadata_container.get_metadata_by_name(col_name)  # pytype: disable=wrong-arg-types  # pandas-drop-duplicates-overloads
    bq_conversion = _TF_INFO_FROM_BQ_DTYPE[metadata.input_data_type]
    feature_output = bq_conversion.series_to_tensor(
        col_data, with_mask=with_mask
    )
    if with_mask and not nested:
      for k, v in feature_output.items():
        output[f'{col_name}_{k}'] = v
    else:
      output[col_name] = feature_output

  return (
      output  # pytype: disable=bad-return-type  # pandas-drop-duplicates-overloads
  )


def _get_output_from_df_iterator(
    output_dataframes: Iterable[pd.DataFrame],
    metadata_container: feature_metadata.FeatureMetadataContainer,
    batch_size: int = 64,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
    drop_remainder: bool = False,
    verbose: bool = False,
) -> Generator[TensorAndMaskType, None, None]:
  """Converts an iterator of DataFrame into the correct format and batch sizes.

  Args:
    output_dataframes: An iterator of pandas DataFrames as generated from
      bigquery.RowIterator.to_dataframe_iterable.
    metadata_container: The metadata generated for the features to be returned
      by the query.
    batch_size: The desired output batch size.
    with_mask: If false the output If true each feature will be returned as a
      dictionary with the keys 'values' and 'was_null'. The 'values' field is a
      dense tensor with null values filled and the 'was_null' tensor is a dense
      boolean tensor that is true where the 'values' tensor was null prior to
      being filled.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.
    drop_remainder: If true no partial batches will be yielded.
    verbose: If true will print debugging messages using tf.print. These can be
      used to help debug when this function is running and when cached data is
      being used.

  Yields:
    A dictionary of features from the query.  If with_mask is true each feature
    will have a dictionary with keys 'values' and 'was_null' where 'values'
    contains a Tensor of the values from that column after null values have been
    filled and 'was_null' is a boolean tensor
  """
  # TODO(b/333154677): Investigate more performant ways to do this.
  data_frame_cache = []
  cur_rows = 0

  conversion_function = functools.partial(
      _dataframe_to_dict_of_tensors,
      metadata_container=metadata_container,
      with_mask=with_mask,
      nested=nested,
  )

  for df in output_dataframes:
    new_rows = df.shape[0]
    cur_rows += new_rows

    # Shortcut the happy path because eventually it might work.
    if new_rows == batch_size and not data_frame_cache:
      yield conversion_function(df)
      cur_rows = 0
      continue

    while cur_rows >= batch_size:
      # We now have enough data to output something
      if data_frame_cache:
        df = pd.concat(data_frame_cache + [df], axis=0, ignore_index=True)
        data_frame_cache = []

      output_data = df.iloc[:batch_size].reset_index(drop=True)

      if cur_rows > batch_size:
        df = df.iloc[batch_size:]
        cur_rows = df.shape[0]
      else:
        cur_rows = 0

      if verbose:
        tf.print('Yielding BQ Query')
      yield conversion_function(output_data)

    if cur_rows > 0:
      data_frame_cache.append(df)

  if data_frame_cache and not drop_remainder:
    if verbose:
      tf.print('Yielding BQ Query')
    yield conversion_function(
        pd.concat(data_frame_cache, axis=0, ignore_index=True)
    )


def bigquery_table_batch_generator(
    metadata_container: feature_metadata.BigQueryTableMetadata,
    bq_client: bigquery.Client,
    bqstorage_client: Optional[bigquery_storage.BigQueryReadClient] = None,
    batch_size: int = 64,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
    drop_remainder: bool = False,
    page_size: Optional[int] = None,
    verbose: bool = False,
) -> Generator[TensorAndMaskType, None, None]:
  """Generates DataFrames iteratively from a Bigquery query.

  Args:
    metadata_container: The metadata generated for the features to be returned
      by the query.
    bq_client: The BigQuery client that will be used for the query.
    bqstorage_client: The BigQueryStorageClient that will be used for the query.
    batch_size: The desired output batch size.
    with_mask: If false the output If true each feature will be returned as a
      dictionary with the keys 'values' and 'was_null'. The 'values' field is a
      dense tensor with null values filled and the 'was_null' tensor is a dense
      boolean tensor that is true where the 'values' tensor was null prior to
      being filled.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.
    drop_remainder: If true no partial batches will be yielded.
    page_size: the pagination size to use when retrieving data from BigQuery. A
      large value can result in fewer BQ calls, hence time savings.
    verbose: If true will print debugging messages using tf.print. These can be
      used to help debug when this function is running and when cached data is
      being used.

  Yields:
    A dictionary of features from the query.  If with_mask is true each feature
    will have a dictionary with keys 'values' and 'was_null' where 'values'
    contains a Tensor of the values from that column after null values have been
    filled and 'was_null' is a boolean tensor
  """
  if page_size is None:
    page_size = batch_size
  row_iterator = bq_client.list_rows(
      metadata_container.bigquery_table,
      selected_fields=metadata_container.to_bigquery_schema(),
      page_size=page_size,
  )

  output_dataframes = row_iterator.to_dataframe_iterable(
      bqstorage_client=bqstorage_client
  )

  yield from _get_output_from_df_iterator(
      output_dataframes=output_dataframes,
      metadata_container=metadata_container,
      batch_size=batch_size,
      with_mask=with_mask,
      nested=nested,
      drop_remainder=drop_remainder,
      verbose=verbose,
  )


def bigquery_query_batch_generator(
    query: str,
    metadata_container: feature_metadata.FeatureMetadataContainer,
    bq_client: bigquery.Client,
    bqstorage_client: Optional[bigquery_storage.BigQueryReadClient] = None,
    batch_size: int = 64,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
    drop_remainder: bool = False,
    page_size: Optional[int] = None,
    verbose: bool = False,
) -> Generator[TensorAndMaskType, None, None]:
  """Generates DataFrames iteratively from a Bigquery query.

  Args:
    query: The query to be run.
    metadata_container: The metadata generated for the features to be returned
      by the query.
    bq_client: The BigQuery client that will be used for the query.
    bqstorage_client: The BigQueryStorageClient that will be used for the query.
    batch_size: The desired output batch size.
    with_mask: If false the output If true each feature will be returned as a
      dictionary with the keys 'values' and 'was_null'. The 'values' field is a
      dense tensor with null values filled and the 'was_null' tensor is a dense
      boolean tensor that is true where the 'values' tensor was null prior to
      being filled.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.
    drop_remainder: If true no partial batches will be yielded.
    page_size: the pagination size to use when retrieving data from BigQuery. A
      large value can result in fewer BQ calls, hence time savings.
    verbose: If true will print debugging messages using tf.print. These can be
      used to help debug when this function is running and when cached data is
      being used.

  Yields:
    A dictionary of features from the query.  If with_mask is true each feature
    will have a dictionary with keys 'values' and 'was_null' where 'values'
    contains a Tensor of the values from that column after null values have been
    filled and 'was_null' is a boolean tensor
  """
  # Page size doesn't seem to have any impact here. Bug in client?
  # Instead, we will handle batching manually which is painful and slow.

  # It is recommended that legacy SQL not be used and this allows large results.
  # For more information see the following documentation:
  #  https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationQuery.FIELDS.allow_large_results
  #  https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationQuery.FIELDS.use_legacy_sql
  query_config = bigquery.QueryJobConfig(use_legacy_sql=_USE_LEGACY_SQL)
  if page_size is None:
    page_size = batch_size
  output_dataframes = (
      bq_client.query(query, job_config=query_config)
      .result(page_size=page_size)
      .to_dataframe_iterable(bqstorage_client=bqstorage_client)
  )

  yield from _get_output_from_df_iterator(
      output_dataframes=output_dataframes,
      metadata_container=metadata_container,
      batch_size=batch_size,
      with_mask=with_mask,
      nested=nested,
      drop_remainder=drop_remainder,
      verbose=verbose,
  )


def tensor_output_signature_from_metadata(
    metadata_container: feature_metadata.FeatureMetadataContainer,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
) -> TensorAndMaskSpecType:
  """Converts the input metadata into a TensorSpec for Dataset creation.

  Args:
    metadata_container: The input feature metadata.
    with_mask: Whether was_null will be included in the output.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.

  Returns:
    The TensorSpec corresponding to the expected output.
  """
  output_spec = {}
  for feature_md in metadata_container:
    if not feature_md.tf_data_type:
      raise ValueError(
          'The TF data type must be set before creating the output signature.'
      )
    name = feature_md.name
    values_spec = tf.TensorSpec(
        (None,), dtype=feature_md.tf_data_type, name=f'{name}_{VALUES_KEY}'
    )
    if with_mask:
      was_null_spec = tf.TensorSpec(
          (None,), dtype=tf.bool, name=f'{feature_md.name}_{WAS_NULL_KEY}'
      )
      if nested:
        output_spec[name] = {
            VALUES_KEY: values_spec,
            WAS_NULL_KEY: was_null_spec,
        }
      else:
        output_spec[f'{name}_{VALUES_KEY}'] = values_spec
        output_spec[f'{name}_{WAS_NULL_KEY}'] = was_null_spec
    else:
      output_spec[name] = values_spec

  return output_spec


def _generate_query_from_metadata(
    table_metadata: feature_metadata.BigQueryTableMetadata,
    limit: Optional[int] = None,
    where_clauses: Sequence[str] = tuple(),
) -> str:
  """Generates a query from column names, where clauses and a limit."""
  select_fields_string = '`,`'.join(col.name for col in table_metadata)
  limit_str = f' LIMIT {limit}' if limit else ''
  where_str = bq_utils.where_statement_from_clauses(where_clauses)
  return (
      f'SELECT `{select_fields_string}` FROM\n'
      f'{table_metadata.escaped_table_id}{where_str}{limit_str}'
  )


def update_tf_data_types_from_bq_data_types(
    metadata_container: feature_metadata.FeatureMetadataContainer,
) -> None:
  """Updates each feature's tf_data_type based on the input_data_type.

  Args:
    metadata_container: The metadata container to be updated.
  """
  for metadata in metadata_container:
    if not metadata.tf_data_type:
      metadata.tf_data_type = _TF_INFO_FROM_BQ_DTYPE[
          metadata.input_data_type
      ].tf_dtype


def get_bigquery_dataset(
    table_metadata: feature_metadata.BigQueryTableMetadata,
    bq_client: bigquery.Client,
    bqstorage_client: Optional[bigquery_storage.BigQueryReadClient] = None,
    batch_size: int = 64,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
    limit: Optional[int] = None,
    cache_location: Optional[str] = NO_CACHE_LOCATION_NAME,
    where_clauses: Sequence[str] = tuple(),
    drop_remainder: bool = False,
    page_size: Optional[int] = None,
) -> tf.data.Dataset:
  """Creates a Big Query dataset for the specified Table.

  Args:
    table_metadata: The metadata for the BigQuery table.
    bq_client: The BigQuery client that will be used for the query.
    bqstorage_client: The BigQueryStorageClient that will be used for the query.
    batch_size: The desired output batch size.
    with_mask: If true each feature will be returned along with a tensor
      indicating if the values were null before being filled.
    nested: If true the output dictionary will have sub-dictionaries each with
      the fields 'values' and 'was_null'. If false the output dictionary will
      have keys '{feature_name}_values' and '{feature_name}_was_null' for each
      feature. Note that this only applies if `with_mask` is true.
    limit: Put a limit on the number of examples returned from the query. If
      falsy all the results will be returned.
    cache_location: If 'disk' the data will be cached to disk if None the data
      will not be cached and otherwise the data will be cached to memory.
    where_clauses: A list of clauses that can be combined with and statements to
      get the correct values in the BigQuery table.
    drop_remainder: If true no partial batches will be yielded.
    page_size: the pagination size to use when retrieving data from BigQuery. A
      large value can result in fewer BQ calls, hence time savings.

  Returns:
    A tf.data.Dataset for the table.
  """
  update_tf_data_types_from_bq_data_types(table_metadata)
  tensor_spec = tensor_output_signature_from_metadata(
      table_metadata, with_mask=with_mask, nested=nested
  )

  if limit or where_clauses:
    query = _generate_query_from_metadata(table_metadata, limit, where_clauses)
    tensor_generator_fn = functools.partial(
        bigquery_query_batch_generator,
        query=query,
        metadata_container=table_metadata,
        bq_client=bq_client,
        bqstorage_client=bqstorage_client,
        batch_size=batch_size,
        with_mask=with_mask,
        nested=nested,
        drop_remainder=drop_remainder,
        page_size=page_size,
    )
  else:
    tensor_generator_fn = functools.partial(
        bigquery_table_batch_generator,
        metadata_container=table_metadata,
        bq_client=bq_client,
        bqstorage_client=bqstorage_client,
        batch_size=batch_size,
        with_mask=with_mask,
        nested=nested,
        drop_remainder=drop_remainder,
        page_size=page_size,
    )

  dataset = tf.data.Dataset.from_generator(
      tensor_generator_fn, output_signature=tensor_spec
  )
  if cache_location is not None and cache_location != NO_CACHE_LOCATION_NAME:
    filename = tempfile.mkdtemp() if cache_location == 'disk' else ''
    # This appears to give a false warning: b/194670791
    dataset = dataset.cache(filename)

  return dataset


def keras_input_from_metadata(
    metadata_container: feature_metadata.FeatureMetadataContainer,
    with_mask: bool = _WITH_MASK_DEFAULT,
    nested: bool = NESTED_FORMAT_DEFAULT,
) -> Mapping[str, Union[Mapping[str, KerasInputType], KerasInputType]]:
  """Creates Keras Input objects for input metadata.

  Args:
    metadata_container: The metadata of the features to be used for input.
    with_mask: Whether a mask is included in the dataset or not. See
      get_bigquery_dataset for more information.
    nested: If True assumes the input dictionaries are nested. Otherwise,
      assumes that a flat dictionary is input. This is only relevant if
      with_mask is true.

  Returns:
    A dictionary of tf.keras.Input objets that can be input into a
    tf.keras.Model object.
  """
  model_input = {}
  for metadata in metadata_container:
    cur_name = metadata.name
    # Keras layers must take in JSON serializable inputs.
    values_input = tf.keras.Input(
        (), name=f'{cur_name}_{VALUES_KEY}', dtype=metadata.tf_data_type_str
    )
    if with_mask:
      was_null_input = tf.keras.Input(
          (), name=f'{cur_name}_{WAS_NULL_KEY}', dtype='bool'
      )
      if nested:
        model_input[cur_name] = {
            VALUES_KEY: values_input,
            WAS_NULL_KEY: was_null_input,
        }
      else:
        model_input[f'{cur_name}_{VALUES_KEY}'] = values_input
        model_input[f'{cur_name}_{WAS_NULL_KEY}'] = was_null_input
    else:
      model_input[cur_name] = values_input
  return model_input


def get_dataset_and_metadata_for_table(
    table_path: Optional[str] = None,
    table_parts: Optional[bq_utils.BQTablePathParts] = None,
    bigquery_client: Optional[bigquery.Client] = None,
    bigquery_storage_client: Optional[
        bigquery_storage.BigQueryReadClient
    ] = None,
    metadata_options: Optional[
        feature_metadata.MetadataRetrievalOptions
    ] = None,
    metadata_builder: Optional[feature_metadata.BigQueryMetadataBuilder] = None,
    batch_size: int = 64,
    with_mask: bool = _WITH_MASK_DEFAULT,
    drop_remainder: bool = False,
    page_size: Optional[int] = None,
) -> Tuple[tf.data.Dataset, feature_metadata.BigQueryTableMetadata]:
  """Gets the metadata and dataset for a BigQuery table.

  Args:
    table_path: The full path ('project.dataset.table') of the BigQuery table.
    table_parts: The parsed potions of the BigQuery table path.
    bigquery_client: The BigQuery Client object to use for getting the metadata.
    bigquery_storage_client: The BigQuery storage client to use for the dataset.
    metadata_options: The metadata retrieval options to use.
    metadata_builder: The metadata builder to use to get the metadata.
    batch_size: The batch size to use for the dataset. Default is 64.
    with_mask: Whether the dataset should be returned with a mask format. For
      more information see get_bigquery_dataset.
    drop_remainder: If true no partial batches will be yielded.
    page_size: the pagination size to use when retrieving data from BigQuery. A
      large value can result in fewer BQ calls, hence time savings.

  Returns:
    A tuple of the output dataset and metadata for the specified table.

  Raises:
    ValueError: If neither table_parts nor table_path are specified or both of
      them are.
  """
  if not (table_parts or table_path):
    raise ValueError('Either table_parts or table_path must be specified.')
  elif table_parts and table_path:
    raise ValueError('Only one of table_parts or table_path can be specified.')

  if not table_parts:
    table_parts = bq_utils.BQTablePathParts.from_full_path(table_path)

  if not bigquery_client:
    bigquery_client = bigquery.Client(project=table_parts.project_id)

  if not bigquery_storage_client:
    bigquery_storage_client = bigquery_storage.BigQueryReadClient()

  if not metadata_options:
    metadata_options = feature_metadata.MetadataRetrievalOptions.get_none()

  if not metadata_builder:
    metadata_builder = (
        feature_metadata.BigQueryMetadataBuilder.from_table_parts(
            table_parts, bq_client=bigquery_client
        )
    )

  all_metadata = metadata_builder.get_metadata_for_all_features(
      metadata_options
  )
  # TODO(b/333154677): Refactor the code so that this extra call is not needed.
  update_tf_data_types_from_bq_data_types(all_metadata)

  dataset = get_bigquery_dataset(
      all_metadata,
      bigquery_client,
      bqstorage_client=bigquery_storage_client,
      batch_size=batch_size,
      with_mask=with_mask,
      # Cache during prepare_dataset instead.
      cache_location=None,
      where_clauses=metadata_options.where_clauses,
      drop_remainder=drop_remainder,
      page_size=page_size,
  )

  return dataset, all_metadata



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class and functions for loading data for the SPADE algorithm.

Includes functionality for loading local CSVs for testing, as well
as functions for interacting with BigQuery where end users read and upload
table data.
"""

import functools
import os
import pathlib
# TODO(b/247116870): Change to collections when Vertex supports python 3.9
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

from absl import logging
from google.cloud import bigquery
from google.cloud import storage
from google.cloud.storage import transfer_manager
import numpy as np
import pandas as pd

from spade_anomaly_detection import parameters
from spade_anomaly_detection.data_utils import bq_dataset
from spade_anomaly_detection.data_utils import bq_utils
from spade_anomaly_detection.data_utils import feature_metadata

import tensorflow as tf

_DATA_ROOT = 'spade_anomaly_detection/example_data/'


def load_dataframe(
    dataset_name: str,
    label_col_index: int = -1,
    filter_label_value: Optional[Union[str, int]] = None,
) -> Sequence[Union[pd.DataFrame, pd.Series]]:
  """Loads csv data located in the ./example_data directory for unit tests.

  Args:
    dataset_name: 'thyroid_labeled' is supported right now.
    label_col_index: Column to use for labels, default to the last column.
    filter_label_value: Value to filter the label column on. Could be a string
      or an integer. If there is no value specified, no filtering will be
      performed.

  Returns:
    A tuple (features, labels), both are dataframes corresponding to the
    features and labels of the dataset respectively.
  """

  file_path = os.path.join(_DATA_ROOT, f'{dataset_name}.csv')

  if dataset_name not in [
      'thyroid_labeled',
      'covertype_pu_labeled',
      'drug_train_pu_labeled',
      'covertype_pnu_10000',
      'covertype_pnu_100000',
  ]:
    raise ValueError(f'Unknown dataset_name: {dataset_name}')

  dataframe = pd.read_csv(file_path, delimiter=',', skiprows=1, index_col=0)

  if len(dataframe.shape) != 2:
    raise ValueError(
        f'{dataset_name} has {len(dataframe.shape)} dimension(s) and is not '
        'equal to the 2 dimensional requirement.'
    )

  if label_col_index >= dataframe.shape[1]:
    raise ValueError(
        f'Label column {label_col_index} is outside of the column range.'
    )

  if filter_label_value is not None:
    dataframe = dataframe[
        dataframe.iloc[:, label_col_index] == filter_label_value
    ]

  features = dataframe.drop(dataframe.columns[label_col_index], axis=1)
  labels = dataframe.iloc[:, label_col_index].astype(int)

  return features, labels


def load_tf_dataset_from_csv(
    dataset_name: str,
    label_col_index: int = -1,
    batch_size: Optional[int] = None,
    filter_label_value: Optional[Any] = None,
    return_feature_count: bool = False,
) -> Union[tf.data.Dataset, Tuple[tf.data.Dataset, int]]:
  """Loads a TensorFlow dataset from the ./example_data directory.

  Args:
    dataset_name: The name of the CSV file to be loaded.
    label_col_index: Column to use for labels, default to the last column.
    batch_size: The number of records the generator returns in a single call.
      Tune this depending on the size of the machines memory the algorithm is
      being executed on. The default value is None, which means that the dataset
      is not batched. In this case, when iterating through the dataset, it will
      yield one record per call instead of a batch of records.
    filter_label_value: Value to filter the label column on. Could be a string
      or an integer.
    return_feature_count: If True, returns a tuple of the Dataset and the number
      of features.

  Returns:
    A TensorFlow dataset, or a tuple containing a Tensorflow Dataset and the
      number of features,
  """
  features, labels = load_dataframe(
      dataset_name, label_col_index, filter_label_value
  )

  feature_tensors = tf.convert_to_tensor(features, dtype=tf.dtypes.float32)

  label_tensors = tf.convert_to_tensor(labels, dtype=tf.dtypes.int8)

  tf_dataset = tf.data.Dataset.from_tensor_slices(
      (feature_tensors, label_tensors)
  )

  if batch_size is not None:
    tf_dataset = tf_dataset.batch(batch_size)

  if return_feature_count:
    return tf_dataset, len(features.columns)
  return tf_dataset


class DataLoader:
  """Contains methods for interacting with BigQuery using RunnerParameters."""

  def __init__(self, runner_parameters: parameters.RunnerParameters):
    self.input_feature_metadata: feature_metadata.BigQueryTableMetadata = None
    self.runner_parameters = runner_parameters

    self.table_parts = bq_utils.BQTablePathParts.from_full_path(
        self.runner_parameters.input_bigquery_table_path
    )

  @property
  def num_features(self) -> int:
    """Returns the number of features in the dataset."""
    if not self.input_feature_metadata:
      raise ValueError(
          'Input feature metadata has not been instantiated yet, '
          'ensure that you have made a call to '
          'load_tf_dataset_from_bigquery() before this method '
          'is called.'
      )
    # Now that the data is loaded, access the metadata and get the number of
    # features. We get this from the number of elements in the metadata sequence
    # and subtracting 1 for the label column.
    return len(self.input_feature_metadata._metadata_sequence) - 1  # pylint:disable=protected-access

  def get_label_thresholds(
      self,
      bigquery_table_path: str,
      where_statements: Optional[List[str]] = None,
  ) -> Mapping[str, float]:
    """Computes positive and negative thresholds based on label ratios.

    This method is useful for setting percentile thresholds when performing
    inference on the OCC ensemble. Feature vectors can be labeled as normal
    or anomalous depending on these values. This method requires that the label
    column, and positive and negative values are set in RunnerParameters.

    Args:
      bigquery_table_path: Full BigQuery table path to compute label ratios on.
      where_statements: A list of valid SQL where statements.

    Returns:
      A dictionary containing 'positive_threshold' and 'negative_threshold'.
    """
    if where_statements is not None:
      additional_where_statements = ' AND ' + ' AND '.join(where_statements)
    else:
      additional_where_statements = ''

    with bigquery.Client(
        project=self.table_parts.project_id
    ) as bigquery_client:
      label_count_query = (
          f'SELECT {self.runner_parameters.label_col_name},'
          f' count({self.runner_parameters.label_col_name}) AS record_count'
          f' FROM {bigquery_table_path}'
          f' WHERE {self.runner_parameters.label_col_name} <>'
          f' {self.runner_parameters.unlabeled_data_value}'
          f' {additional_where_statements}'
          f' GROUP BY {self.runner_parameters.label_col_name}'
      )
      # Note: Parameterization was explored, however there were severe
      # limitations including the inability to parameterize the table name and
      # the unlabeled data value. The latter is due to the fact that the label
      # column name and the label vale can be strings or integers, and in order
      # to make SQL comparisons, these must be the same type.
      label_count_query_job = bigquery_client.query(query=label_count_query)

    label_counts = label_count_query_job.result().to_dataframe()
    logging.info('Label counts: %s', label_counts)

    positive_count = label_counts[
        label_counts[self.runner_parameters.label_col_name]
        == self.runner_parameters.positive_data_value
    ].iat[0, -1]

    labeled_data_record_count = label_counts['record_count'].sum()

    positive_threshold = 100 * (positive_count / labeled_data_record_count)
    label_thresholds = {
        'positive_threshold': positive_threshold,
        'negative_threshold': 100 - positive_threshold,
    }

    if self.runner_parameters.verbose:
      logging.info('Computed label thresholds: %s', label_thresholds)

    return label_thresholds

  def get_query_record_result_length(
      self,
      input_path: str,
      where_statements: Optional[List[str]] = None,
  ) -> int:
    """Returns an integer representing the total number of rows in the query.

    Args:
      input_path: BigQuery string in the format 'project.dataset.table'.
      where_statements: A list of valid SQL where statements.

    Returns:
      An integer representing the number of rows in the query.
    """

    with bigquery.Client(
        project=self.table_parts.project_id
    ) as big_query_client:
      query = f'SELECT count(*) FROM `{input_path}` '
      if where_statements:
        query += 'WHERE ' + ' AND '.join(where_statements)

      query_job = big_query_client.query(query)
      result = query_job.result().to_dataframe().values

    return int(result)

  def _preprocess(
      self,
      batch: Mapping[str, tf.Tensor],
      label_col_name: str,
      convert_label_to_int: bool = False,
      convert_features_to_float64: bool = False,
  ) -> Tuple[tf.Tensor, tf.Tensor]:
    """Reshapes the results returned by the dataloader utility.

    Method that is used as a mapping function on a TensorFlow dataset. Instead
    of returning a single matrix, we adjust the results to be two matrices of
    (M, Nx), (M, 1). Where M is the numnber of training examples and Nx is the
    dimension of the feature vector.

    Args:
      batch: A batch of tensors in the format of (key, value) where the key is a
        string defingin the column name in BigQuery.
      label_col_name: The name of the label column in the input BigQuery table.
      convert_label_to_int: Set to True to cast the contents of the label column
        to integer.
      convert_features_to_float64: Set to True to cast the contents of the
        features columns to float64.

    Returns:
      A tuple of Tensors in the format (features, labels)
    """
    feature_matrix = []
    label_tensor = None

    for key, value in batch.items():
      if key == label_col_name:
        label_tensor = value
        continue

      feature_matrix.append(value)

    if label_tensor is None:
      raise ValueError('Label column not found in the input table.')
    if convert_label_to_int:
      label_tensor = tf.cast(label_tensor, tf.int64)

    feature_matrix = tf.stack(feature_matrix, axis=1)
    if convert_features_to_float64:
      feature_matrix = tf.cast(feature_matrix, tf.float64)

    return feature_matrix, label_tensor

  def load_tf_dataset_from_bigquery(
      self,
      input_path: str,
      label_col_name: str,
      where_statements: Optional[List[str]] = None,
      ignore_columns: Optional[Sequence[str]] = None,
      batch_size: Optional[int] = None,
      label_column_filter_value: Optional[int | list[int]] = None,
      convert_features_to_float64: bool = False,
      page_size: Optional[int] = None,
  ) -> tf.data.Dataset:
    """Loads a TensorFlow dataset from a BigQuery Table.

    Args:
      input_path: BigQuery string in the format 'project.dataset.table'.
      label_col_name: The name of the label column in the input BigQuery table.
      where_statements: A sequence of SQL where clauses that you wish to apply
        when selecting data from BigQuery.
      ignore_columns: A sequence of columns that will not be included in the
        returned dataset. Default is None, meaning all columns are included.
      batch_size: The number of records the generator returns in a single call.
        Tune this depending on the size of the machines memory the algorithm is
        being executed on. The default value is None, which means that the
        dataset is not batched. In this case, when iterating through the
        dataset, it will yield one record per call instead of a batch of
        records.
      label_column_filter_value: An integer or list of integers used when
        filtering the label column values. No value will result in all data
        returned from the table.
      convert_features_to_float64: Set to True to cast the contents of the
        features columns to float64.
      page_size: the pagination size to use when retrieving data from BigQuery.
        A large value can result in fewer BQ calls, hence time savings.

    Returns:
      A TensorFlow dataset.
    """
    metadata_retrieval_options = None

    metadata_builder = None

    if label_column_filter_value is not None:
      where_statements = (
          list() if where_statements is None else where_statements
      )
      if isinstance(label_column_filter_value, int):
        where_statements.append(
            f'{label_col_name} = {label_column_filter_value}'
        )
      else:
        where_statements.append(
            f'CAST({label_col_name} AS INT64) IN '
            f'UNNEST({label_column_filter_value})'
        )

    if ignore_columns is not None:
      metadata_builder = feature_metadata.BigQueryMetadataBuilder(
          project_id=self.table_parts.project_id,
          bq_dataset_name=self.table_parts.bq_dataset_name,
          bq_table_name=self.table_parts.bq_table_name,
          ignore_columns=ignore_columns,
      )

    if where_statements:
      metadata_retrieval_options = (
          feature_metadata.MetadataRetrievalOptions.get_none(
              where_clauses=where_statements
          )
      )

    tf_dataset, metadata = bq_dataset.get_dataset_and_metadata_for_table(
        table_path=input_path,
        batch_size=batch_size,
        with_mask=False,
        drop_remainder=True,
        metadata_options=metadata_retrieval_options,
        metadata_builder=metadata_builder,
        page_size=page_size,
    )
    options = tf.data.Options()
    # Avoid a large warning output by TF Dataset.
    options.deterministic = False
    tf_dataset = tf_dataset.with_options(options)

    self.input_feature_metadata = metadata

    preprocess_fn = functools.partial(
        self._preprocess,
        label_col_name=label_col_name,
        convert_label_to_int=False,
        convert_features_to_float64=convert_features_to_float64,
    )
    return tf_dataset.map(
        lambda batch: preprocess_fn(batch),  # pylint:disable=unnecessary-lambda
        num_parallel_calls=tf.data.AUTOTUNE,
    )

  def construct_schema_from_dataframe(
      self, df: pd.DataFrame
  ) -> Sequence[bigquery.SchemaField]:
    """Constructs a BigQuery schema from a Pandas dataframe."""

    def get_type(dtype: np.dtype) -> str:
      if dtype in [np.float32, np.float64]:
        return 'FLOAT'
      elif dtype in [np.int32, np.int64, np.int_]:
        return 'INTEGER'
      elif dtype in [np.bool, np.bool_]:
        return 'BOOLEAN'
      else:
        raise ValueError(f'Unsupported data type: {dtype}')

    columns = df.columns
    dtypes = df.dtypes
    return [
        bigquery.SchemaField(
            name=col, field_type=get_type(dtype), mode='NULLABLE'
        )
        for (col, dtype) in zip(columns, dtypes)
    ]

  def construct_dataframe_from_features(
      self,
      features: np.ndarray,
      labels: np.ndarray,
      convert_label_to_bool: bool = False,
  ) -> pd.DataFrame:
    """Constructs a dataframe from features and labels.

    Args:
      features: The features as a 2D Numpy array.
      labels: The labels as a 1D Numpy array.
      convert_label_to_bool: If True, convert the label to boolean.

    Returns:
      A dataframe containing the features and labels.
    """
    if not self.input_feature_metadata:
      raise ValueError(
          'Input feature metadata has not been instantiated yet, '
          'ensure that you have made a call to '
          'load_tf_dataset_from_bigquery() before this method '
          'is called.'
      )
    combined_data = np.concatenate(
        [features, labels.reshape(len(features), 1)], axis=1
    )

    column_names = list(self.input_feature_metadata.names)
    column_names.remove(self.runner_parameters.label_col_name)
    column_names.append(self.runner_parameters.label_col_name)

    complete_dataframe = pd.DataFrame(data=combined_data, columns=column_names)

    if convert_label_to_bool:
      # Adjust label column type so that users can go straight to BigQuery or
      # AutoML without having to adjust data. Both of these products require a
      # boolean or string target column, not integer.
      complete_dataframe[self.runner_parameters.label_col_name] = (
          complete_dataframe[self.runner_parameters.label_col_name].astype(
              'bool'
          )
      )
    return complete_dataframe

  def parse_gcs_folder_uri(self, gcs_uri: str) -> tuple[str, str]:
    """Parses a GCS URI of a folder."""
    # Get ID of GCS bucket from URI. The URIs are of the form
    # 'gs://bucket-name/folder/path'.
    gcs_prefix = 'gs://'
    uri_without_prefix = gcs_uri.split(gcs_prefix)[-1]
    bucket_name = uri_without_prefix.partition('/')[0]
    folder_name = uri_without_prefix.partition('/')[-1]
    return bucket_name, folder_name

  def upload_data_to_local_disk(
      self,
      file_path: str,
      features: np.ndarray,
      labels: np.ndarray,
      convert_label_to_bool: bool = False,
      use_parquet: bool = True,
  ):
    """Saves a batch of pseudo-labeled data to local disk on the VM."""
    complete_dataframe = self.construct_dataframe_from_features(
        features=features,
        labels=labels,
        convert_label_to_bool=convert_label_to_bool,
    )
    if use_parquet:
      # Send the dataframe to the file. Compression can be 'snappy', 'gzip',
      # or 'lz4'. Anecdotally, 'gzip' is 4X slower than 'snappy' or 'lz4'.
      complete_dataframe.to_parquet(
          file_path, index=False, compression='snappy'
      )
    else:
      # Send the dataframe to the file.
      complete_dataframe.to_csv(file_path, header=True, index=False)

  def _construct_file_paths_for_gcs(
      self,
      source_folder: str,
  ) -> Sequence[str]:
    """Constructs the file paths for uploading to GCS.

    Args:
      source_folder: Name of local folder from where files are to be uploaded.

    Returns:
      List of file paths of files in the source folder/
    """
    # Generate a list of paths (in string form) relative to the `directory`.
    # This can be done in a single list comprehension, but is expanded into
    # multiple lines here for clarity.

    # First, recursively get all files in `directory` as Path objects.
    directory_as_path_obj = pathlib.Path(source_folder)
    paths = directory_as_path_obj.rglob('*')

    # Filter so the list only includes files, not directories themselves.
    file_paths = [path for path in paths if path.is_file()]

    # These paths are relative to the current working directory. Next, make them
    # relative to `directory`
    relative_paths = [path.relative_to(source_folder) for path in file_paths]

    # Finally, convert them all to strings.
    string_paths = [str(path) for path in relative_paths]
    return string_paths

  def upload_folder_to_gcs(
      self,
      source_folder: str,
      gcs_uri: str,
      num_workers: int = 8,
      use_parquet: bool = True,
  ):
    """Uploads all the files in a local folder to GCS."""
    file_ext = 'parquet' if use_parquet else 'csv'

    # Parse the GCS URI. The URIs are of the form
    # 'gs://bucket-name/folder/path'.
    bucket_name, folder_name = self.parse_gcs_folder_uri(gcs_uri)

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)

    # Make the folders on GCS if they don't exist.
    if not tf.io.gfile.exists(gcs_uri):
      tf.io.gfile.makedirs(gcs_uri)

    # Generate a list of paths (in string form) relative to the `directory`.
    string_paths = self._construct_file_paths_for_gcs(source_folder)
    logging.info('Found %d %s files.', len(string_paths), file_ext)

    # Start the upload.
    results = transfer_manager.upload_many_from_filenames(
        bucket,
        string_paths,
        blob_name_prefix=f'{folder_name}/',
        source_directory=source_folder,
        max_workers=num_workers,
    )

    upload_count = 0
    for name, result in zip(string_paths, results):
      # The results list is either `None` or an exception for each filename in
      # the input list, in order.
      if isinstance(result, Exception):
        logging.error('Failed to upload %s due to exception: %s', name, result)
      else:
        upload_count += 1
    logging.info(
        'Uploaded %d %s files to %s in folder %s.',
        upload_count,
        file_ext,
        bucket.name,
        folder_name,
    )

  def _get_bigquery_job_config(
      self,
      schema: Sequence[bigquery.SchemaField],
      write_disposition: Optional[str] = 'WRITE_TRUNCATE',
      use_parquet: bool = True,
  ) -> bigquery.LoadJobConfig:
    """Gets the BigQuery job configuration."""
    source_format = (
        bigquery.SourceFormat.PARQUET
        if use_parquet
        else bigquery.SourceFormat.CSV
    )
    if use_parquet:
      job_config = bigquery.LoadJobConfig(
          schema=schema,
          source_format=source_format,
          write_disposition=write_disposition,
      )
    else:
      job_config = bigquery.LoadJobConfig(
          schema=schema,
          source_format=source_format,
          skip_leading_rows=1,  # Skip the header row.
          write_disposition=write_disposition,
      )
    return job_config

  def load_bigquery_table_from_gcs_uri(
      self,
      gcs_uri: str,
      table_name: str,
      schema: Sequence[bigquery.SchemaField],
      write_disposition: Optional[str] = 'WRITE_TRUNCATE',
      use_parquet: bool = True,
  ) -> None:
    """Loads a BigQuery table from a GCS URI.

    CSV load job quotas are here:
    https://cloud.google.com/bigquery/quotas#load_jobs

    Args:
      gcs_uri: GCS URI (potentially with wildcard) from which to load data.
      table_name: Name of target BigQuery table.
      schema: Schema of the target BigQuery table.
      write_disposition: Whether to overwrite the table (WRITE_TRUNCATE) or
        append to the table (WRITE_APPEND). Default is `WRITE_TRUNCATE`.
      use_parquet: If True, use parquet files on GCS, else use CSV files.
    """
    with bigquery.Client(
        project=self.table_parts.project_id
    ) as big_query_client:
      job_config = self._get_bigquery_job_config(
          schema=schema,
          write_disposition=write_disposition,
          use_parquet=use_parquet,
      )
      file_ext = 'parquet' if use_parquet else 'csv'
      load_job = big_query_client.load_table_from_uri(
          f'{gcs_uri}/*.{file_ext}', table_name, job_config=job_config
      )
      load_job.result()  # Waits for the job to complete.

    destination_table = big_query_client.get_table(table_name)
    logging.info(
        'Loaded %d rows from %s files to table %s.',
        destination_table.num_rows,
        file_ext,
        table_name,
    )

  def upload_dataframe_as_bigquery_table(
      self,
      features: np.ndarray,
      labels: np.ndarray,
  ) -> None:
    """Uploads the dataframe to BigQuery, create or replace table.

    Args:
      features: Numpy array of features.
      labels: Numpy array of labels.
    """
    if not self.input_feature_metadata:
      raise ValueError(
          'Input feature metadata has not been instantiated yet, '
          'ensure that you have made a call to '
          'load_tf_dataset_from_bigquery() before this method '
          'is called.'
      )
    combined_data = np.concatenate(
        [features, labels.reshape(len(features), 1)], axis=1
    )

    column_names = list(self.input_feature_metadata.names)
    column_names.remove(self.runner_parameters.label_col_name)
    column_names.append(self.runner_parameters.label_col_name)

    complete_dataframe = pd.DataFrame(data=combined_data, columns=column_names)

    # Adjust label column type so that users can go straight to BigQuery or
    # AutoML without having to adjust data. Both of these products require a
    # boolean or string target column, not integer.
    complete_dataframe[self.runner_parameters.label_col_name] = (
        complete_dataframe[self.runner_parameters.label_col_name].astype('bool')
    )

    with bigquery.Client(
        project=self.table_parts.project_id
    ) as big_query_client:
      job_config = bigquery.LoadJobConfig(
          # Creates or replaces an existing table.
          write_disposition='WRITE_TRUNCATE',
      )
      # In this configuration, the schema will be auto-detected.
      big_query_client.load_table_from_dataframe(
          dataframe=complete_dataframe,
          destination=self.runner_parameters.output_bigquery_table_path,
          job_config=job_config,
      )
    if self.runner_parameters.verbose:
      logging.info(
          'Uploaded pseudo-labeled data to %s',
          self.runner_parameters.output_bigquery_table_path,
      )

  def get_metadata_for_table(
      self,
      bigquery_client: Optional[bigquery.Client] = None,
      metadata_options: Optional[
          feature_metadata.MetadataRetrievalOptions
      ] = None,
      metadata_builder: Optional[
          feature_metadata.BigQueryMetadataBuilder
      ] = None,
  ) -> feature_metadata.BigQueryTableMetadata:
    """Gets the metadata and dataset for a BigQuery table.

    Args:
      bigquery_client: The BigQuery Client object to use for getting the
        metadata.
      metadata_options: The metadata retrieval options to use.
      metadata_builder: The metadata builder to use to get the metadata.

    Returns:
      The metadata for the specified table.
    """
    if not bigquery_client:
      bigquery_client = bigquery.Client(project=self.table_parts.project_id)

    if not metadata_options:
      metadata_options = feature_metadata.MetadataRetrievalOptions.get_none()

    if not metadata_builder:
      metadata_builder = (
          feature_metadata.BigQueryMetadataBuilder.from_table_parts(
              self.table_parts, bq_client=bigquery_client
          )
      )

    all_metadata = metadata_builder.get_metadata_for_all_features(
        metadata_options
    )
    # TODO(b/333154677): Refactor code so that this extra call is not needed.
    bq_dataset.update_tf_data_types_from_bq_data_types(all_metadata)

    return all_metadata



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilities for interfacing with BigQuery."""

import dataclasses
import enum
import re
from typing import Optional, Mapping, Sequence


@dataclasses.dataclass
class BQTablePathParts:
  """Data object with the parts of a full BigQuery table.

  Attributes:
    project_id: The project id for the project that the BigQuery table is in.
    bq_dataset_name: The name of the dataset where the table resides.
    bq_table_name: The table's name.
  """
  project_id: str
  bq_dataset_name: str
  bq_table_name: str

  @property
  def full_table_id(self) -> str:
    """The full path 'project.dataset.table' of the BigQuery table."""
    return f"{self.project_id}.{self.bq_dataset_name}.{self.bq_table_name}"

  @property
  def escaped_table_id(self) -> str:
    """The full path of the BigQuery table escaped with `."""
    return f"`{self.full_table_id}`"

  @classmethod
  def from_full_path(cls, full_big_query_path: str) -> "BQTablePathParts":
    """Parses a full BigQuery path into components (project, dataset, table).

    Args:
      full_big_query_path: The full path (e.g.
        project_id.dataset_name.table_name) to a BigQuery table.

    Returns:
      A BQTablePathParts object.
    """
    # TODO(b/333154677): Update to handle optional project.
    project_prefix_pattern = re.compile(
        r"`?(?P<project_id>[^.`]+)\."
        r"(?P<bq_dataset_name>[^.`]+)"
        r"(?:$|\.(?P<bq_table_name>[^.`]+)`?$)",
        re.VERBOSE,
    )

    path_match = project_prefix_pattern.match(full_big_query_path)
    if not path_match:
      raise ValueError(
          f"The BigQuery table path {full_big_query_path} did not have the "
          f"expected format of project_id.dataset_name.table_name")

    output_dict = path_match.groupdict()
    for name, match in output_dict.items():
      if not match:
        raise ValueError(
            f"The {name} was not found in the BQ path {full_big_query_path}")

    # Note that the regex match names must match the attribute names.
    return cls(
        project_id=output_dict["project_id"],
        bq_dataset_name=output_dict["bq_dataset_name"],
        bq_table_name=output_dict["bq_table_name"],
    )


def where_statement_from_clauses(where_clauses: Sequence[str],
                                 conjunction: str = "AND") -> str:
  """Constructs an optional where statement from list of conditions.

  Args:
    where_clauses: A sequence of strings each of which is a valid clause for the
      table being queried.
    conjunction: The conjunction used to join the clauses. Does not need to
      include spaces as they will be added later.

  Returns:
    Either an empty string if there were no clauses or a where clause with
    whitespace on both sides.
  """
  if not where_clauses:
    return ""

  join_str = f") {conjunction} ("
  combined_clauses = f"({join_str.join(where_clauses)})"
  return f"\nWHERE {combined_clauses} "


@enum.unique
class SplitColumnValues(enum.Enum):
  TRAIN = enum.auto()
  VALIDATE = enum.auto()
  TEST = enum.auto()


def create_split_column_conditions(
    split_column: Optional[str],) -> Mapping[SplitColumnValues, Sequence[str]]:
  return {
      mode: (f"{split_column} = '{mode.name}'",) if split_column else ()
      for mode in SplitColumnValues
  }



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates feature metadata for use in normalization and feature creation."""
import collections
import dataclasses
import enum
import logging
import textwrap
import typing
# TODO(b/249435057): Convert imports to use collections.abc once the container
#  is updated to have python 3.9.
from typing import (
    Any,
    DefaultDict,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Union,
)

from google.cloud import bigquery
import numpy as np
import tensorflow as tf

# pylint: disable=g-statement-before-imports,g-import-not-at-top
if typing.TYPE_CHECKING:
  from spade_anomaly_detection.data_utils import bq_utils
else:
  try:
    from spade_anomaly_detection.data_utils import bq_utils
  except ImportError:
    import bq_utils
# pylint: enable=g-statement-before-imports,g-import-not-at-top


# https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
_BIGQUERY_SUPPORTED_NUMERIC_DATA_TYPES = frozenset((
    'INT64',
    'NUMERIC',
    'BIGNUMERIC',
    'FLOAT64',
    'FLOAT',
    'INTEGER',
))

# TODO(b/333154677): Decide if we should handle dates and times as numerical as
# well.
# TODO(b/333154677): Research if there are other data types we should try to
# handle.
_BIGQUERY_SUPPORTED_DISCRETE_DATA_TYPES = frozenset((
    'STRING',
    'BYTES',
    'BOOL',
    'BOOLEAN',
    'INT64',
    'INTEGER',
    'DATE',
    'DATETIME',
    'TIMESTAMP',
    'TIME',
))
_BIGQUERY_SUPPORTED_DATA_TYPES = frozenset((
    _BIGQUERY_SUPPORTED_NUMERIC_DATA_TYPES
    | _BIGQUERY_SUPPORTED_DISCRETE_DATA_TYPES
))


_USE_LEGACY_SQL = False


@enum.unique
class Normalization(str, enum.Enum):
  """GAIN normalization methods."""

  MINMAX = 'minmax'
  STANDARD = 'standard'


@dataclasses.dataclass
class MetadataRetrievalOptions:
  """Describes what metadata should be collected."""

  # For numeric features
  get_mean: bool = True
  get_variance: bool = True
  get_min: bool = True
  get_max: bool = True
  get_median: bool = False
  get_log_mean: bool = True
  get_log_variance: bool = True
  number_of_quantiles: Optional[int] = None

  # Transform related parameters.
  min_log_value: float = 1.0e-4

  # For discrete features
  get_mode: bool = True
  max_vocab_size: int = 5000

  # BigQuery specific options.
  where_clauses: Sequence[str] = dataclasses.field(default_factory=tuple)

  def no_options(self, ignored_fields: Optional[Set[str]] = None) -> bool:
    """True if metadata statistics do not need to be retrieved.

    Args:
      ignored_fields: Any fields to not check for a boolean true value. If None
        is passed in the default of {'min_log_value'} will be used.

    Returns:
      True if none of the important attributes are true.
    """
    if ignored_fields is None:
      ignored_fields = {'min_log_value', 'where_clauses'}

    for field in dataclasses.fields(self):
      if field.name in ignored_fields:
        continue

      if getattr(self, field.name):
        return False

    return True

  @classmethod
  def get_all(
      cls,
      number_of_quantiles: int = 101,
      max_vocab_size: int = 10000,
      where_clauses: Sequence[str] = tuple(),
  ) -> 'MetadataRetrievalOptions':
    """Creates a MetadataRetrievalOptions object with all values True.

    Args:
      number_of_quantiles: The number of quantiles to use.
      max_vocab_size: The maximum vocab size to use.
      where_clauses: A list of clauses that can be combined with and statements
        to get the correct values in the BigQuery table.

    Returns:
      A MetadataRetrievalOptions that gets all metadata classes.
    """
    return cls(
        get_mean=True,
        get_variance=True,
        get_min=True,
        get_max=True,
        get_median=True,
        get_log_mean=True,
        get_log_variance=True,
        number_of_quantiles=number_of_quantiles,
        get_mode=True,
        max_vocab_size=max_vocab_size,
        where_clauses=where_clauses,
    )

  @classmethod
  def get_none(
      cls,
      where_clauses: Sequence[str] = tuple(),
  ) -> 'MetadataRetrievalOptions':
    """Creates a MetadataRetrievalOptions object that doesn't get any metadata.

    Args:
      where_clauses: A list of clauses that can be combined with and statements
        to get the correct values in the BigQuery table.

    Returns:
      A MetadataRetrievalOptions that gets no metadata.
    """
    return cls(
        get_mean=False,
        get_variance=False,
        get_min=False,
        get_max=False,
        get_median=False,
        get_log_mean=False,
        get_log_variance=False,
        min_log_value=0.0,
        number_of_quantiles=None,
        get_mode=False,
        max_vocab_size=0,
        where_clauses=where_clauses,
    )

  @classmethod
  def for_normalization(
      cls,
      normalization: Normalization,
      where_clauses: Sequence[str] = tuple(),
  ) -> 'MetadataRetrievalOptions':
    """Returns the metadata retrieval options needed for a normalization.

    Args:
      normalization: The normalization method to be used. Currently, supports
        minmax and standard.
      where_clauses: A list of clauses that can be combined with and statements
        to get the correct values in the BigQuery table.

    Returns:
      A MetadataRetrievalOptions with the necessary options being True.

    Raises:
      ValueError: If an invalid normalization is passed in.
    """
    if normalization == Normalization.MINMAX:
      return cls(
          get_mean=False,
          get_variance=False,
          get_min=True,
          get_max=True,
          get_median=False,
          get_log_mean=False,
          get_log_variance=False,
          get_mode=False,
          where_clauses=where_clauses,
      )
    elif normalization == Normalization.STANDARD:
      return cls(
          get_mean=True,
          get_variance=True,
          get_min=False,
          get_max=False,
          get_median=False,
          get_log_mean=False,
          get_log_variance=False,
          get_mode=False,
          where_clauses=where_clauses,
      )
    else:
      raise ValueError(f'The normalization {normalization} is not valid.')


# noinspection PyUnresolvedReferences
@dataclasses.dataclass
class FeatureMetadata:
  """The metadata for a single data column.

  Attributes:
    name: The name of the column.
    index: The index of the column (e.g. 0-N-1).
    input_data_type: The data_type of the column in BigQuery
    tf_data_type_str: The data type that should be used for the feature in
      TensorFlow as a string.
    mean: The mean of the feature. For numeric data only.
    variance: The variance of the feature. For numeric data only.
    min: The minimum value of the feature. For numeric data only.
    max: The maximum value of the feature. For numeric data only.
    median: The mean of the feature. For numeric data only.
    log_shift: The value to shift the data by before applying a log transform.
      This is effectively MIN(0, eps - min(X)). Therefore the log transform
      should be applied as log(X + log_shift). Will be present if log_mean or
      log_variance are present. For numeric data only.
    log_mean: The mean of the feature after the log transform. For numeric data
      only.
    log_variance: The variance of the feature after the log transform. For
      numeric data only.
    quantiles: A list of quantiles evenly spaced between 0% (the min) and 100%
      (the max). As an example if there are three values they are the min, the
      median, and the max.
    cardinality: The number of different values the feature takes on in the
      table. For discrete data only.
    mode: The most common value of the feature. For discrete data only.
    vocabulary: A mapping between the feature values and the number of times
      that they occur in the data. Note that this may not be present for some
      discrete data whose cardinality is too large. For discrete data only.
    is_numeric: True if the BigQuery datatype is numeric.
    is_discrete: True it fhe BigQuery datatype is discrete.
  """

  name: str
  index: int
  input_data_type: Optional[str] = None
  tf_data_type_str: Optional[str] = None
  # These are related to numeric features.
  mean: Optional[float] = None
  variance: Optional[float] = None
  min: Optional[float] = None
  max: Optional[float] = None
  median: Optional[float] = None
  log_shift: Optional[float] = None
  log_mean: Optional[float] = None
  log_variance: Optional[float] = None
  quantiles: Optional[Sequence[float]] = None
  # These are related to discrete features.
  # TODO(b/333154677): Consider if we should calculate cardinality for all
  # features.
  #  as sometimes float fields can only take on a set of discrete values.
  cardinality: Optional[int] = None
  mode: Optional[Any] = None
  vocabulary: Optional[Mapping[Any, int]] = None

  @property
  def tf_data_type(self) -> Optional[tf.dtypes.DType]:
    """The TensorFlor data type of the feature."""
    return (
        tf.dtypes.as_dtype(self.tf_data_type_str)
        if self.tf_data_type_str
        else None
    )

  @tf_data_type.setter
  def tf_data_type(self, tf_data_type: Union[tf.dtypes.DType, np.dtype, str]):
    tf_dtype = tf.dtypes.as_dtype(tf_data_type)
    self.tf_data_type_str = tf_dtype.name

  @property
  def range(self) -> Optional[float]:
    """The range (max-min) of the data."""
    if self.max is None or self.min is None:
      return None

    return self.max - self.min

  @property
  def is_numeric(self) -> bool:
    """If the feature is numeric."""
    if self.tf_data_type_str:
      return self.tf_data_type.is_floating or self.tf_data_type.is_integer  # pytype: disable=attribute-error  # always-use-return-annotations
    # TODO(b/333154677): Handle other input sources more generically.
    else:
      return self.input_data_type in _BIGQUERY_SUPPORTED_NUMERIC_DATA_TYPES

  @property
  def is_discrete(self) -> bool:
    """If the feature is discrete."""
    tf_data_type = self.tf_data_type
    if tf_data_type:
      return (
          tf_data_type.is_integer
          or tf_data_type.is_bool
          or tf_data_type == tf.dtypes.string
      )
    # TODO(b/333154677): Handle other input sources more generically.
    else:
      return self.input_data_type in _BIGQUERY_SUPPORTED_DISCRETE_DATA_TYPES

  def get_config(self) -> MutableMapping[str, Any]:
    """Creates a config that can be converted to json."""
    config = dataclasses.asdict(self)
    # Because JSON only takes strings as keys we need to track and convert
    # types.
    if self.vocabulary:
      # TODO(b/333154677): Consider whether we should limit the vocabulary to
      # one type.
      config['vocabulary_types'] = {
          v: type(v).__name__ for v in self.vocabulary
      }
    return config

  @classmethod
  def from_config(cls, config: MutableMapping[str, Any]) -> 'FeatureMetadata':
    """Creates a FeatureMetadata from a config."""

    def _update_vocabulary_key(new_key, old_key):
      config['vocabulary'][new_key] = config['vocabulary'][old_key]
      del config['vocabulary'][old_key]

    vocabulary_types = config.pop('vocabulary_types', {})
    for k, k_type in vocabulary_types.items():
      if k_type == 'bool':
        bool_key = k.lower() == 'true'
        _update_vocabulary_key(bool_key, k)
      elif k_type == 'NoneType':
        _update_vocabulary_key(None, k)
      elif k_type == 'int':
        _update_vocabulary_key(int(k), k)
      elif k_type == 'float':
        _update_vocabulary_key(float(k), k)

    return cls(**config)


MetadataDict = Optional[
    Union[
        collections.defaultdict[Optional[Any], Any],
        DefaultDict[str, MutableSequence[FeatureMetadata]],
    ]
]


class FeatureMetadataContainer:
  """Allows feature metadata to be easily retrieved.

  Example::

    # Pass retrieved metadata into the container.
    builder = BigQueryMetadataBuilder('project', 'dataset', 'table')
    all_metadata = builder.get_feature_names_and_types()
    metadata_container = FeatureMetadataContainer(all_metadata)

    # Get a feature by name.
    feature1_metadata = metadata_container.get_metadata_by_name('feature1')
    assert feature1_metadata.name == 'feature1'

    # Get features by dtype.
    float_features = metadata_container.get_metadata_by_name('FLOAT64')
    assert all([ff.input_data_type == 'FLOAT64' for ff in float_features])

    # Get a feature by index and check it matches.
    assert metadata_container[0] == all_metadata[0]

    # Iterate through the metadata and check it matches.
    for idx, c_metadata in enumerate(metadata_container):
      assert c_metadata == all_metadata[idx]
  """

  def __init__(
      self,
      all_feature_metadata: Union[
          Sequence[FeatureMetadata], 'FeatureMetadataContainer'
      ],
  ):
    """Container that holds feature metadata for easy retrieval.

    Args:
      all_feature_metadata: A sequence of the metadata for the features in the
        table.
    """
    # Internal storage of the input metadata as a sequence.
    self._metadata_sequence = (
        all_feature_metadata._metadata_sequence
        if isinstance(all_feature_metadata, FeatureMetadataContainer)
        else all_feature_metadata
    )

    # Lazily created cache for mapping output features to data types.
    self._metadata_for_dtype: Optional[
        DefaultDict[str, MutableSequence[FeatureMetadata]]
    ] = None
    # Lazily created cache for quickly getting feature metadata by name.
    self._metadata_for_name: Optional[Mapping[str, FeatureMetadata]] = None

  @property
  def feature_metadata_by_names(self) -> Mapping[str, FeatureMetadata]:
    """A mapping between the name and metadata of each feature.

    Raises:
      ValueError: If two features have the same name.
    """
    if not self._metadata_for_name:
      self._metadata_for_name = {m.name: m for m in self._metadata_sequence}
      # Make sure all names are unique
      if len(self._metadata_for_name) != len(self._metadata_sequence):
        raise ValueError('Two features may not have the same name.')

    return self._metadata_for_name

  def get_metadata_by_name(self, feature_name: str) -> FeatureMetadata:
    """Returns the metadata for the feature with the given name.

    Args:
      feature_name: The name of the feature (i.e. metadata.name).

    Returns:
      The metadata object whose name matches the input name.
    """
    return self.feature_metadata_by_names[feature_name]

  def __getitem__(self, idx: int) -> FeatureMetadata:
    """Allow user to use access the metadata sequence directly.

    Args:
      idx: The index (based on input order) of the feature to retrieve.

    Returns:
      The metadata for the specified feature.
    """
    return self._metadata_sequence[idx]

  @property
  def feature_metadata_by_dtypes(
      self,
  ) -> MetadataDict:
    """A mapping between  data type and metadata for all features."""
    if not self._metadata_for_dtype:
      self._metadata_for_dtype = collections.defaultdict(list)
      for m in self._metadata_sequence:
        self._metadata_for_dtype[m.input_data_type].append(m)

    return self._metadata_for_dtype

  def get_metadata_by_dtype(
      self, input_data_type: str
  ) -> MutableSequence[FeatureMetadata]:
    """Returns a set of metadata for features of the given type."""
    return self.feature_metadata_by_dtypes[input_data_type]

  @property
  def names(self) -> Sequence[str]:
    return tuple(self.feature_metadata_by_names.keys())

  def __iter__(self):
    """Iterates through the metadata in the order they were provided."""
    for metadata in self._metadata_sequence:
      yield metadata

  def __len__(self) -> int:
    """Returns the number of elements in the metadata."""
    return len(self._metadata_sequence)

  def __repr__(self):
    metadata_str = ','.join([f'{f!r}' for f in self])
    return f'{type(self)!r}(all_feature_metadata=[{metadata_str}])'

  def to_bigquery_schema(self) -> MutableSequence[bigquery.SchemaField]:
    """Creates a BigQuery schema from the feature metadata.

    Returns:
      A list of the SchemaFields for each of the features.
    """
    # TODO(b/333154677): Add in the nullable attribute.
    return [bigquery.SchemaField(m.name, m.input_data_type) for m in self]

  def equal_names_and_types(
      self,
      other: 'FeatureMetadataContainer',
      difference_method: Optional[str] = None,
  ) -> bool:
    """Returns where the name and datatypes of the two containers is equal.

    Args:
      other: Another metadata container to compare to.
      difference_method: How to handle differences. Valid values are 'raise',
        logger levels or None. For logging levels see:
        https://docs.python.org/3/library/logging.html#logging-levels.

    Returns:
      True if all the names and types of the two schema match.

    Raises:
      ValueError: If the difference method is raise and the two values do not
        match or an invalid difference_method is specified.
    """
    fields_to_check = ('name', 'input_data_type', 'tf_data_type')

    if difference_method == 'raise':

      def difference_fn(message):
        raise ValueError(message)

    elif difference_method:
      numeric_level = getattr(logging, difference_method.upper(), None)
      if numeric_level is None:
        raise ValueError(
            f'The difference_method {difference_method} is not valid.'
        )

      def difference_fn(message):
        logging.log(numeric_level, message)

    else:

      def difference_fn(_):
        pass

    if len(self) != len(other):
      difference_fn(
          f'The length of the two containers is different: {self} vs {other}'
      )
      return False

    different_features = []
    for feature, other_feature in zip(self, other):
      for field_name in fields_to_check:
        if getattr(feature, field_name) != getattr(other_feature, field_name):
          different_features.append((feature, other_feature))
          continue

    if different_features:
      features_str = [
          f'{mine} vs {other}' for mine, other in different_features
      ]
      difference_fn(f'The following features differ: {";".join(features_str)}')
      return False

    return True

  def get_config(self) -> MutableMapping[str, Any]:
    """Creates a config that can be converted to json."""
    return {
        'all_feature_metadata': [
            fm.get_config() for fm in self._metadata_sequence
        ],
    }

  @classmethod
  def from_config(
      cls,
      config: MutableMapping[str, Any],
  ) -> 'FeatureMetadataContainer':
    """Create a FeatureMetadataContainer from a config."""
    if 'all_feature_metadata' not in config:
      raise ValueError('The key all_feature_metadata must be in the config.')

    config['all_feature_metadata'] = [
        FeatureMetadata.from_config(fm) for fm in config['all_feature_metadata']
    ]
    return cls(**config)


class BigQueryTableMetadata(FeatureMetadataContainer):
  """Contains the metadata for features in a BigQuery table.

  This extends the normal FeatureMetadataContainer by including information
  about the datasource (i.e. the BigQuery Table).

  Attributes:
    project_id: The name of the GCP project that the BigQuery table is in.
    bq_dataset_name: The name of the dataset that the BigQuery table is in.
    bq_table_name: The name of the table from which the metadata was retrieved.
  """

  def __init__(
      self,
      all_feature_metadata: Sequence[FeatureMetadata],
      project_id: str,
      bq_dataset_name: str,
      bq_table_name: str,
  ):
    """Initializer for a BigQueryTableMetadata object.

    Args:
      all_feature_metadata: Sequence of FeatureMetadata for the columns.
      project_id: The name of the GCP project that the BigQuery table is in.
      bq_dataset_name: The name of the dataset that the BigQuery table is in.
      bq_table_name: The name of the table from which the metadata was
        retrieved.
    """
    super().__init__(all_feature_metadata)

    self.project_id = project_id
    self.bq_dataset_name = bq_dataset_name
    self.bq_table_name = bq_table_name

  @property
  def full_table_id(self) -> str:
    """The full path 'project.dataset.table' of the BigQuery table."""
    return f'{self.project_id}.{self.bq_dataset_name}.{self.bq_table_name}'

  @property
  def escaped_table_id(self) -> str:
    """The full path of the BigQuery table escaped with `."""
    return f'`{self.full_table_id}`'

  @property
  def bigquery_table(self) -> bigquery.Table:
    """The bigquery.Table object for this table."""
    return bigquery.Table.from_string(self.full_table_id)

  def update_bq_path_parts(self, bq_file_parts: bq_utils.BQTablePathParts):
    self.project_id = bq_file_parts.project_id
    self.bq_dataset_name = bq_file_parts.bq_dataset_name
    self.bq_table_name = bq_file_parts.bq_table_name

  def __repr__(self):
    metadata_str = ','.join([f'{f!r}' for f in self])
    return (
        f'{type(self)!r}(project_id={self.project_id}, '
        f'bq_dataset_name={self.bq_dataset_name}, '
        f'bq_table_name={self.bq_table_name}, '
        f'all_feature_metadata=[{metadata_str}])'
    )

  def get_config(self) -> MutableMapping[str, Any]:
    """Creates a config that can be converted to json."""
    config = super().get_config()
    config.update({
        'project_id': self.project_id,
        'bq_dataset_name': self.bq_dataset_name,
        'bq_table_name': self.bq_table_name,
    })
    return config


class BigQueryMetadataBuilder:
  """Gets metadata about the column in a BQ table."""

  def __init__(
      self,
      project_id: str,
      bq_dataset_name: str,
      bq_table_name: str,
      ignore_columns: Iterable[str] = (),
      bq_client: Optional[bigquery.Client] = None,
  ):
    """Creates an object that can retrieve metadata about the input BQ table.

    Args:
      project_id: The BigQuery project that the table is in.
      bq_dataset_name: The BigQuery dataset that has the table.
      bq_table_name: The name of the table itself.
      ignore_columns: Any columns that should not be included in the output.
      bq_client: A BigQuery client to be used to interact with BigQuery. If this
        is not provided (default) a new client will be created for the input
        project.
    """
    self._project_id = project_id
    self._bq_dataset_name = bq_dataset_name
    self._bq_table_name = bq_table_name
    self._ignore_columns = ignore_columns
    self._bq_client = bq_client or bigquery.Client(project=project_id)
    self._query_config = bigquery.QueryJobConfig(use_legacy_sql=_USE_LEGACY_SQL)

    # Set up initial values for private variables that will be lazily updated
    # and cached.
    self._rows = None

  @property
  def full_table_id(self) -> str:
    """The full path 'project.dataset.table' of the BigQuery table."""
    return f'{self._project_id}.{self._bq_dataset_name}.{self._bq_table_name}'

  @property
  def escaped_table_id(self) -> str:
    """The full path of the BigQuery table escaped with `."""
    return f'`{self.full_table_id}`'

  def _query_bq(
      self,
      query: str,
      job_config: Optional[bigquery.QueryJobConfig] = None,
      **kwargs,
  ) -> bigquery.QueryJob:
    job_config = job_config or self._query_config
    return self._bq_client.query(query, job_config=job_config, **kwargs)

  @property
  def rows(self) -> int:
    if not self._rows:
      row_query = textwrap.dedent(
          f"""\
          SELECT COUNT(*)
          FROM {self.escaped_table_id}"""
      )
      self._rows = next(self._query_bq(row_query).result())[0]

    return self._rows

  def get_feature_names_and_types(self) -> BigQueryTableMetadata:
    """Returns a metadata collection of all the columns in the table.

    Returns:
      A BigQueryTableMetadata with FeatureMetadata for each column in the
      object's BQ table.
    """
    table = self._bq_client.get_table(self.full_table_id)

    if self._ignore_columns:
      ignored_set = set(self._ignore_columns)
    else:
      ignored_set = set()

    columns = []
    for idx, column in enumerate(table.schema):
      if column.name not in ignored_set:
        if column.field_type not in _BIGQUERY_SUPPORTED_DATA_TYPES:
          raise NotImplementedError(
              f'The datatype {column.field_type} for {column.name} '
              'is not currently supported.'
          )
        columns.append(FeatureMetadata(column.name, idx, column.field_type))

    return BigQueryTableMetadata(
        columns,
        project_id=self._project_id,
        bq_dataset_name=self._bq_dataset_name,
        bq_table_name=self._bq_table_name,
    )

  def _construct_numeric_metadata_query(
      self, feature: FeatureMetadata, options: MetadataRetrievalOptions
  ) -> str:
    """Constructs a query to get the numeric metadata for the input feature.

    Note that the suffix of each of the query results must match the attribute
    that it will be assigned to.

    In our tests doing this as a single query for each feature appeared to give
    the best performance as it avoided multiple queries but also avoided the
    anti-pattern of throwing a bunch of queries together:
    https://cloud.google.com/bigquery/docs/best-practices-performance-compute#split_complex_queries_into_multiple_smaller_ones


    Args:
      feature: The current metadata for the feature. Must include the name.
      options: Specifications for what types of metadata should be retrieved.

    Returns:
      The query to run in bigquery to get the metadata.
    """
    if not feature:
      raise ValueError('feature cannot be None.')
    if not feature.input_data_type:
      raise ValueError('feature input_data_type cannot be None.')
    if not feature.is_numeric:
      raise ValueError(f'This function only works for numeric data: {feature}')

    # Cache for brevity, performance and ease of use.
    name = feature.name
    # The CTE queries list can contain multiple CTE expression which will be
    # proceeded by the WITH statement and joined with commas.
    cte_queries = []

    # The where clauses list will be proceeded by the WHERE statement and
    # joined with ands.
    where_clauses = list(options.where_clauses)

    # The select components will be combined with commas to create the main
    # query. The output of this query is what will be used to create the
    # results.
    main_query_components = []
    if options.get_mean:
      main_query_components.append(f'AVG({name}) as {name}_mean')

    if options.get_variance:
      main_query_components.append(f'VARIANCE({name}) as {name}_variance')

    if options.get_min:
      main_query_components.append(f'MIN({name}) as {name}_min')

    if options.get_max:
      main_query_components.append(f'MAX({name}) as {name}_max')

    if options.get_median:
      # TODO(b/333154677): Look at other ways of doing this.
      main_query_components.append(
          f'APPROX_QUANTILES({name}, 3)[OFFSET(1)] as {name}_median'
      )

    if options.number_of_quantiles:
      main_query_components.append(
          f'APPROX_QUANTILES({name}, {options.number_of_quantiles}) '
          f'as {name}_quantiles'
      )

    is_float = feature.input_data_type.startswith('FLOAT')

    if options.get_log_mean or options.get_log_variance:
      cte_where = (
          where_clauses + [f'NOT IS_NAN({name})'] if is_float else where_clauses
      )
      cte_shift = 'log_shift'
      where_statement = bq_utils.where_statement_from_clauses(cte_where)
      cte_queries.append(
          f'{cte_shift} AS ('
          f'SELECT {options.min_log_value} '
          f'- LEAST(0, MIN({name})) as value '
          f'FROM {self.escaped_table_id}{where_statement})'
      )
      log_shift_subquery = f'(SELECT value from {cte_shift})'

      cte_transformed = 'log_transformed'
      cte_queries.append(
          f'{cte_transformed} AS ('
          f'SELECT LOG({name} + {log_shift_subquery}) as value '
          f'FROM {self.escaped_table_id}{where_statement})'
      )

      main_query_components.append(f'{log_shift_subquery} as {name}_log_shift')
      if options.get_log_mean:
        main_query_components.append(
            f'(SELECT AVG(value) FROM {cte_transformed}) as {name}_log_mean'
        )
      if options.get_log_variance:
        main_query_components.append(
            '(SELECT VARIANCE(value) '
            f'FROM {cte_transformed}) as {name}_log_variance'
        )

    query_parts = []
    if cte_queries:
      combined_with = ',\n'.join(cte_queries)
      query_parts.append(f'WITH {combined_with}')

    combined_select = ',\n  '.join(main_query_components)
    if is_float:
      where_clauses.append(f'NOT IS_NAN({name})')

    where_statement = bq_utils.where_statement_from_clauses(where_clauses)
    query_parts.append(
        f'SELECT\n{combined_select}\n'
        f'FROM {self.escaped_table_id}{where_statement}'
    )

    return '\n'.join(query_parts)

  def update_numeric_feature_metadata(
      self, feature: FeatureMetadata, options: MetadataRetrievalOptions
  ) -> None:
    """Updates a numeric feature's metadata based on BigQuery results.

    Args:
      feature: The feature to update.
      options: The options for getting the metadata.
    """
    query_string = self._construct_numeric_metadata_query(feature, options)
    # We only expect one row of results.
    query_result = next(self._query_bq(query_string).result())
    if not query_result:
      raise ValueError(f'Expected a result from query: {query_string}')

    # TODO(b/333154677): Do this more efficiently and safely by only iterating
    # over numeric attributes.
    query_columns = set(query_result.keys())
    for metadata_field in dataclasses.fields(feature):
      field_name = metadata_field.name
      # This means we overwrite any already existing data.
      # TODO(b/333154677): Figure out a cleaner way to do this.
      query_result_name = f'{feature.name}_{field_name}'
      if query_result_name in query_columns:
        if query_result[query_result_name] is None:
          logging.warning('Got a None value for %s on %s', field_name, feature)
        setattr(feature, field_name, query_result[query_result_name])

  def update_discrete_feature_metadata(
      self, feature: FeatureMetadata, options: MetadataRetrievalOptions
  ) -> None:
    """Updates a discrete feature's metadata based on BigQuery results.

    Args:
      feature: The feature to update.
      options: The options for getting the metadata.
    """
    if not feature.is_discrete:
      raise ValueError(f'This function only works for discrete data: {feature}')

    # Cache for brevity, performance and ease of use.
    name = feature.name

    where_clauses = [f'{name} IS NOT NULL']
    where_clauses.extend(options.where_clauses)

    # We will get the mode from the first APPROX_TOP_COUNT output
    num_top_count = max(options.max_vocab_size, int(options.get_mode))

    select_parts = [
        # Should be close enough to use APPROX_COUNT_DISTINCT instead.
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/approximate_aggregate_functions#approx_count_distinct
        f'APPROX_COUNT_DISTINCT({name}) as {name}_cardinality',
    ]
    if num_top_count <= 10000:
      # While it is not documented APPROX_TOP_COUNT does not appear to work
      # with numbers >10000. When this occurs we have to use 2 queries
      # instead.
      select_parts.append(
          f'APPROX_TOP_COUNT({name}, {num_top_count}) as {name}_top_count'
      )

    combined_select = ',\n  '.join(select_parts)
    # The where clauses list will be proceeded by the WHERE statement and
    # joined with ands.
    where_statement = bq_utils.where_statement_from_clauses(where_clauses)
    query_string = (
        f'SELECT\n  {combined_select}\n'
        f'FROM {self.escaped_table_id}{where_statement}'
    )

    # We only expect one row of results.
    query_result = next(self._query_bq(query_string).result())
    if not query_result:
      raise ValueError(f'Expected a result from query: {query_string}')

    # We need to handle mode independently since it comes from the vocab.
    feature.cardinality = int(query_result[f'{feature.name}_cardinality'])

    if num_top_count <= 10000:
      feature.vocabulary = {
          item['value']: int(item['count'])
          for item in query_result[f'{feature.name}_top_count']
      }
    else:
      # Since we are asking for a vocabulary of over 10k elements we need to use
      # a second query to get those results and the convert it to a dictionary.
      dataframe = (
          self._query_bq(
              f'SELECT {name} as value, count(*) as count '
              f'FROM {self.escaped_table_id}{where_statement} GROUP BY {name} '
              f'ORDER BY 2 DESC LIMIT {num_top_count}'
          )
          .result()
          .to_dataframe()
      )
      feature.vocabulary = {
          r['value']: int(r['count']) for r in dataframe.to_dict('records')
      }

    # The vocabulary is already ordered by the top count.
    if options.get_mode:
      feature.mode = next(iter(feature.vocabulary))

  def get_metadata_for_all_features(
      self, options: MetadataRetrievalOptions
  ) -> BigQueryTableMetadata:
    """Gets the metadata_collection for all the features in the specified table.

    Args:
      options: The options to use when retrieving the metadata.

    Returns:
      A collection of metadata for each feature in the table with the specified
      values calculated.
    """
    feature_metadata = self.get_feature_names_and_types()

    if options.no_options():
      return feature_metadata

    for feature in feature_metadata:
      if feature.is_numeric:
        self.update_numeric_feature_metadata(feature, options)

      if feature.is_discrete:
        self.update_discrete_feature_metadata(feature, options)

    return feature_metadata

  # TODO(b/333154677): Add the ability to export the metadata_collection.

  @classmethod
  def from_table_parts(
      cls, table_parts: bq_utils.BQTablePathParts, *args, **kwargs
  ) -> 'BigQueryMetadataBuilder':
    """Constructs a BigQueryMetadataBuilder directly from table_parts.

    Args:
      table_parts: The BigQuery path parts for the table for the metadata.
      *args: Positional arguments to be passed to __init__.
      **kwargs: Keyword arguments to be passed to __init__.

    Returns:
      The BigQueryMetadataBuilder for the specified table.
    """
    return cls(
        project_id=table_parts.project_id,
        bq_dataset_name=table_parts.bq_dataset_name,
        bq_table_name=table_parts.bq_table_name,
        *args,
        **kwargs,
    )

  @classmethod
  def from_table_path(
      cls, table_path: str, *args, **kwargs
  ) -> 'BigQueryMetadataBuilder':
    """Constructs a BigQueryMetadataBuilder directly from a full table path.

    Args:
      table_path: The full path of the BigQuery table in the format
        project.dataset.table.
      *args: Positional arguments to be passed to __init__.
      **kwargs: Keyword arguments to be passed to __init__.

    Returns:
      The BigQueryMetadataBuilder for the specified table.
    """
    table_parts = bq_utils.BQTablePathParts.from_full_path(table_path)
    return cls.from_table_parts(table_parts, *args, **kwargs)



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for helping to run tests with bq_dataset."""
from typing import Sequence

import pandas as pd


def mock_get_bigquery_dataset_return_value(
    mock_client: ...,
    return_value: Sequence[pd.DataFrame],
    mock_list_rows: bool = True,
    mock_query: bool = True,
) -> None:
  """A utility to help mock out bq_dataset return values.

  Args:
    mock_client: A mock of a bigquery.Client instance. This can be created with:
      mock.create_autospec(bigquery.Client, instance=True, spec_set=True).
    return_value: The return value
    mock_list_rows:
    mock_query:
  """
  if mock_query:
    (
        mock_client.query.return_value.result.return_value.to_dataframe_iterable.return_value
    ) = return_value
  if mock_list_rows:
    (mock_client.list_rows.return_value.to_dataframe_iterable.return_value) = (
        return_value
    )



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Supervised models for SPADE."""
import abc
import contextlib
import dataclasses
from typing import Any, Optional

from absl import logging
import numpy as np
import tensorflow as tf
import tensorflow_decision_forests as tfdf


# TODO(b/247116870): Add models such as MLP so that we can update the parameters
# incrementally via batches of data. TFDF does not support batch training.
class Model(abc.ABC):
  """Abstract class to ensure a consistent API surface for future models."""

  def __init__(self, parameters: Any) -> None:
    self.supervised_parameters = parameters
    self.supervised_model = None

  @abc.abstractmethod
  def train(
      self,
      features: np.ndarray,
      labels: np.ndarray,
      weights: Optional[np.ndarray],
  ) -> tf.keras.callbacks.History:
    pass

  def save(self, save_location: str) -> None:
    """Saves TensorFlow model assets to a GCS location.

    Args:
      save_location: String denoting a Google Cloud Storage location, or local
        disk path. Note that local assets will be deleted when the VM running
        this container is shutdown at the end of the training job.
    """
    if self.supervised_model is not None:
      self.supervised_model.save(save_location)
      logging.info('Saved model assets to %s', save_location)
    else:
      logging.warning('No model to save.')


@dataclasses.dataclass
class RandomForestParameters:
  """Dataclass for the RandomForest TF model.

  Attributes:
    task: Keras task - I have defaulted to classification since we are
      predicting a discrete label.
    num_trees: The number of individual decision trees in the forest.
    min_examples: Minimum number of examples per node.
    max_num_nodes: Maximum number of nodes in a tree. There is no max as a
      default.
    max_depth: Maximum depth of each tree, a depth of 1 would make all nodes
      roots for example.
    random_seed: Random seed for training the model. Learners are deterministic
      when you pass in the same seed.
  """

  task: tfdf.keras.Task = tfdf.keras.Task.CLASSIFICATION
  num_trees: int = 300
  min_examples: int = 5
  max_num_nodes: Optional[int] = None
  max_depth: int = 16
  random_seed: int = 0


class RandomForestModel(Model):
  """RandomForest implemented in TensorFlow."""

  def __init__(
      self,
      parameters: RandomForestParameters,
      distribution_strategy: Optional[tf.distribute.Strategy] = None,
  ):
    """Initializes a random forest model.

    Args:
      parameters: An initialized RandomForestParameters object set with the
        desired values for a RandomForest model.
      distribution_strategy: TensorFlow distribution strategy to use when
        training the random forest model. This is useful on larger datasets,
        where training time is not satisfactory. Refer to the Vertex Custom
        training jobs and TensorFlow distribution strategy for choosing the
        right method here. Default is no strategy, ie use a single worker node.
        It may be worthwhile to attempt to scale up before scaling out, since
        there is typically higher overhead when scaling out.
    """

    super().__init__(parameters)

    # Note on performance: As of tfdf 1.0.0, hardware accelerators are not yet
    # supported. CPU performance for training and inference should be sufficient
    # for most use cases.
    with (
        distribution_strategy.scope()
        if distribution_strategy
        else contextlib.nullcontext()
    ):
      self.supervised_model = tfdf.keras.RandomForestModel(
          **dataclasses.asdict(self.supervised_parameters)
      )
      self.supervised_model.compile(
          metrics=[
              tf.keras.metrics.AUC(name='Supervised_Model_AUC'),
              tf.keras.metrics.Precision(
                  thresholds=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
                  name='Supervised_Model_Precision',
              ),
              tf.keras.metrics.Recall(
                  thresholds=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
                  name='Supervised_Model_Recall',
              ),
          ]
      )

  def train(
      self,
      features: np.ndarray,
      labels: np.ndarray,
      weights: Optional[np.ndarray] = None,
  ) -> tf.keras.callbacks.History:
    """Trains a random forest model.

    Args:
      features: Numpy array of input features - numerical and scaled values
        only.
      labels: Array of labels corresponding to rows of features. The numerical
        label we are trying to predict for the given task.
      weights: Array of sample weights containing values between 0 and 1, where
        1 has the most influence and 0 the least influence on training. These
        values also correspond to rows of features.

    Returns:
      The callback history from the call to fit. This can be useful for
      inspecting loss and model convergence.
    """
    # Note: Batch fitting is not supported in TFDF Random Forest, even though
    # it is in the public documentation.
    return self.supervised_model.fit(
        x=features,
        y=labels,
        sample_weight=weights,
    )



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the entry point for the Docker container to execute the SPADE algo.

See spade/scripts for examples on launching a Vertex training job and using this
script as an entry point.
"""
import logging
from typing import Sequence

from absl import app
from absl import flags

from spade_anomaly_detection import parameters
from spade_anomaly_detection import runner

_TRAIN_SETTING = flags.DEFINE_enum_class(
    "train_setting",
    default="PNU",
    required=False,
    enum_class=parameters.TrainSetting,
    help=(
        "The 'PNU' setting will train the supervised model using ground truth"
        " negative, positive data, and pseudo labeled positive and negative"
        " data. The 'PU' setting will only use negative data from the pseudo"
        " labeler along with the rest of the positive data (ground truth plus"
        " pseudo labeled) to train the supervised model. For model"
        " evaluation,we still require ground truth negative data to be in the"
        " BigQuery dataset, it just won't be used during training."
    ),
)

_INPUT_BIGQUERY_TABLE_PATH = flags.DEFINE_string(
    "input_bigquery_table_path",
    default=None,
    required=True,
    help=(
        "BigQuery table path used for training the anomaly detection model."
        " This needs to be the format 'project.dataset.table'."
    ),
)

_OUTPUT_GCS_URI = flags.DEFINE_string(
    "output_gcs_uri",
    default=None,
    required=True,
    help=(
        "Cloud Storage location to store the supervised model assets. The"
        "location should be in the form gs://bucketname/foldername. A timestamp"
        "will be added to the end of the folder so that multiple runs of this"
        "won't overwrite previous runs."
    ),
)

_LABEL_COL_NAME = flags.DEFINE_string(
    "label_col_name",
    default=None,
    required=True,
    help="The name of the label column in BigQuery.",
)

_POSITIVE_DATA_VALUE = flags.DEFINE_integer(
    "positive_data_value",
    default=None,
    required=True,
    help="The column value used to define an anomalous (positive) data point.",
)

_NEGATIVE_DATA_VALUE = flags.DEFINE_integer(
    "negative_data_value",
    default=None,
    required=True,
    help="The column value used to define a normal (negative) data point.",
)

_UNLABELED_DATA_VALUE = flags.DEFINE_integer(
    "unlabeled_data_value",
    default=None,
    required=True,
    help="The column value used to define an unlabeled data point.",
)

_POSITIVE_THRESHOLD = flags.DEFINE_float(
    "positive_threshold",
    default=None,
    required=False,
    help=(
        "Integer used as the percentile for the one class classifier ensemble "
        "to label a point as positive. The closer to 0 this value is set, the "
        "less positive data will be labeled. However, we expect an increase in "
        "precision when lowering the value, and an increase in recall when "
        "raising it. Equavalent to saying the given data point needs to be "
        "located in the top X percentile in order to be considered anomalous."
    ),
)

_NEGATIVE_THRESHOLD = flags.DEFINE_float(
    "negative_threshold",
    default=None,
    required=False,
    help=(
        "Integer used as the percentile for the one class classifier ensemble"
        " to label a point as negative. The higher this value is set, the less"
        " negative data will be labeled. A value in the range of 50-99 is a"
        " good starting point. We expect an increase in precision when raising"
        " this value, and an increase in recall when lowering it. Equavalent to"
        " saying the given data point needs to be X percentile or greater in"
        " order to be considered anomalous."
    ),
)

_IGNORE_COLUMNS = flags.DEFINE_list(
    "ignore_columns",
    default=None,
    required=False,
    help=(
        "A list of columns located in the input table that you would "
        "like to ignore in both pseudo labeling and supervised model training."
    ),
)

_WHERE_STAEMENTS = flags.DEFINE_list(
    "where_statements",
    default=None,
    required=False,
    help=(
        "Additional SQL where statements with correct syntax that can be "
        "leveraged to filter BigQuery table data. An example is "
        "\"WHERE date > '2008-11-11'\"."
    ),
)

_TEST_BIGQUERY_TABLE_PATH = flags.DEFINE_string(
    "test_bigquery_table_path",
    default=None,
    required=False,
    help=(
        "A complete BigQuery path in the form of 'project.dataset.table' to be "
        "used for evaluating the supervised model. Note that the positive and "
        "negative label values must also be the same in this testing set. It "
        "is okay to to have your test labels in that form, or use 1 for "
        "positive and 0 for negative."
    ),
)

_TEST_LABEL_COL_NAME = flags.DEFINE_string(
    "test_label_col_name",
    default=None,
    required=False,
    help="The label column name in the test dataset.",
)

_TEST_DATASET_HOLDOUT_FRACTION = flags.DEFINE_float(
    "test_dataset_holdout_fraction",
    default=0.3,
    required=False,
    help=(
        "Float between 0 and 1 representing the fraction of samples to hold "
        "out as a test set."
    ),
)

_OUTPUT_BIGQUERY_TABLE_PATH = flags.DEFINE_string(
    "output_bigquery_table_path",
    default=None,
    required=False,
    help=(
        "A complete BigQuery path in the form of 'project.dataset.table' to be"
        " used for uploading the pseudo labeled data. This includes features"
        " and new labels. By default, we will use the column names from the"
        " input_bigquery_table_path BigQuery table."
    ),
)

_ALPHA = flags.DEFINE_float(
    "alpha",
    default=1.0,
    required=False,
    help=(
        "Sample weights for weighting the loss function, only for"
        " pseudo-labeled data from the occ ensemble. Original data that is"
        " labeled will have a weight of 1.0."
    ),
)

_BATCHES_PER_MODEL = flags.DEFINE_integer(
    "batches_per_model",
    default=1,
    required=False,
    help=(
        "The number of batches to use when fitting a single model in the"
        " ensemble. Default is one, meaning that the dataset is divided into"
        " 1/ensemble_count pieces that are used to fit different models."
        " Increasing this value is useful for large datasets that can not fit a"
        " 1/ensemble_count shard in memory."
    ),
)

_MAX_OCC_BATCH_SIZE = flags.DEFINE_integer(
    "max_occ_batch_size",
    default=50000,
    required=False,
    help=(
        "The maximum number of examples in a batch to use when fitting a single"
        "occ model"
    ),
)

_BATCH_SIZE = flags.DEFINE_integer(
    "labeling_and_model_training_batch_size",
    default=None,
    required=False,
    help=(
        "The number of examples to use when pseudo labeling unlabeled training"
        " examples. The subset of labeled examples are then fed to the"
        " supervised model to perform a train step. When there is no batch size"
        " specified here, we will use the entire dataset to score, label, and"
        " train the supervised model. For large datasets, use the"
        " batches_per_model and this setting to reduce the amount of data"
        " stored locally on the training machine."
    ),
)

_ENSEMBLE_COUNT = flags.DEFINE_integer(
    "ensemble_count",
    default=5,
    required=False,
    help=(
        "The number of models to use in the labeling ensemble. The default is"
        " 5, and the more that are added to the ensemble decreases the"
        " probability of all of them agreeing on negative and positive samples."
    ),
)

_VERBOSE = flags.DEFINE_bool(
    "verbose",
    default=False,
    required=False,
    help=(
        "The amount of console logs to display during training. Use False to "
        "show few messages, and True for displaying many aspects of model "
        "training and scoring. This is useful for debugging model performance."
    ),
)

_UPLOAD_ONLY = flags.DEFINE_bool(
    "upload_only",
    default=False,
    required=False,
    help=(
        "Use this setting in conjunction with output_bigquery_table_path. When"
        " True, the algorithm will just upload the pseudo labeled data to the"
        " specified table, and will skip training a supervised model. When set"
        " to False, the algorithm will also train a supervised model and upload"
        " to a GCS endpoint. Default is False."
    ),
)

# TODO(b/247116870) Implement the rest of the input parameters.


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  runner_parameters = parameters.RunnerParameters(
      train_setting=_TRAIN_SETTING.value,
      input_bigquery_table_path=_INPUT_BIGQUERY_TABLE_PATH.value,
      output_gcs_uri=_OUTPUT_GCS_URI.value,
      label_col_name=_LABEL_COL_NAME.value,
      positive_data_value=_POSITIVE_DATA_VALUE.value,
      negative_data_value=_NEGATIVE_DATA_VALUE.value,
      unlabeled_data_value=_UNLABELED_DATA_VALUE.value,
      positive_threshold=_POSITIVE_THRESHOLD.value,
      negative_threshold=_NEGATIVE_THRESHOLD.value,
      ignore_columns=_IGNORE_COLUMNS.value,
      where_statements=_WHERE_STAEMENTS.value,
      test_bigquery_table_path=_TEST_BIGQUERY_TABLE_PATH.value,
      test_label_col_name=_TEST_LABEL_COL_NAME.value,
      test_dataset_holdout_fraction=_TEST_DATASET_HOLDOUT_FRACTION.value,
      upload_only=_UPLOAD_ONLY.value,
      output_bigquery_table_path=_OUTPUT_BIGQUERY_TABLE_PATH.value,
      alpha=_ALPHA.value,
      batches_per_model=_BATCHES_PER_MODEL.value,
      max_occ_batch_size=_MAX_OCC_BATCH_SIZE.value,
      labeling_and_model_training_batch_size=_BATCH_SIZE.value,
      ensemble_count=_ENSEMBLE_COUNT.value,
      verbose=_VERBOSE.value,
  )
  runner_obj = runner.Runner(runner_parameters)
  runner_obj.run()


if __name__ == "__main__":
  try:
    app.run(main)
  except Exception as e:
    logging.shutdown()
    raise e



# Copyright 2024 The spade_anomaly_detection Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
# Copyright 2024 The SPADE Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Executes the semi-supervised framework with user defined options.

Runs the implementation of the SPADE framework, which is focused on
anomaly detection. In general, there is a semi-supervised step where we use
a series of one class classifiers to label unlabled data and then use the new
data set to batch fit a supervised model.
"""

# TODO(b/247116870): Change to collections when Vertex supports python 3.9
from typing import Optional, Tuple

from absl import logging
import numpy as np
import pandas as pd

from spade_anomaly_detection import data_loader
from spade_anomaly_detection import occ_ensemble
from spade_anomaly_detection import parameters
from spade_anomaly_detection import supervised_model


class Runner:
  """Runs the SPADE algorithm with values in the RunnerParameters instance.

  Attributes:
    runner_parameters: An instance of the runner_parameters dataclass. This
      holds all parameter information for running a training job.
    test_x: Evaluation features, this is None when there is no evaluation set.
    test_y: Evaluation features, this is None when there is no evaluation set.
    supervised_model_object: None when upload only is True, otherwise we
      instantiate a supervised model object to train on after pseudo labeling.
    supervised_model_metrics: Metrics for the supervised model to be evaluated
      on. By default, this is AUC.
    data_loader: An instance of the data loader, useful for performing a number
      of operations such as loading, filtering, and uploading of bigquery data.
  """

  def __init__(self, runner_parameters: parameters.RunnerParameters):
    self.runner_parameters = runner_parameters
    logging.info('Runner parameters: %s', self.runner_parameters)

    self.data_loader = data_loader.DataLoader(self.runner_parameters)

    # TODO(b/247116870): Evaluate performance implications of using a global
    # testing array - the machine may not have enough memory to store the test
    # set in addition to iterating over the training set.
    self.test_x: Optional[np.ndarray] = None
    self.test_y: Optional[np.ndarray] = None

    if not self.runner_parameters.upload_only:
      self.supervised_model_metrics: Optional[dict[str, float]] = None
      supervised_model_parameters = supervised_model.RandomForestParameters()
      self.supervised_model_object = supervised_model.RandomForestModel(
          supervised_model_parameters
      )
    else:
      self.supervised_model_object = None

    if (
        self.runner_parameters.positive_threshold is None
        or self.runner_parameters.negative_threshold is None
    ):
      input_table_statistics = self.data_loader.get_label_thresholds(
          self.runner_parameters.input_bigquery_table_path
      )

      self.runner_parameters.positive_threshold = (
          input_table_statistics['positive_threshold']
          if self.runner_parameters.positive_threshold is None
          else self.runner_parameters.positive_threshold
      )
      self.runner_parameters.negative_threshold = (
          input_table_statistics['negative_threshold']
          if self.runner_parameters.negative_threshold is None
          else self.runner_parameters.negative_threshold
      )

  def check_data_tables(
      self,
      total_record_count: int,
      unlabeled_record_count: int,
  ) -> None:
    """Runs sanity checks on the table that is passed to this algorithm.

    Args:
      total_record_count: Count of all records in the BigQuery table.
      unlabeled_record_count: Number of unlabeled records in the table.
    """
    if not total_record_count:
      raise ValueError(
          'There are no records in the table: '
          f'{self.runner_parameters.input_bigquery_table_path}'
      )
    elif total_record_count < self.runner_parameters.ensemble_count:
      raise ValueError(
          'There are not enough records in the table to fit one '
          'record per model in the ensemble. '
          f'Total records: {total_record_count}'
      )

    if total_record_count < 1000:
      logging.warning(
          (
              'Using a small number of examples to train the model, results '
              'will vary significantly between runs. Total records: '
              ' %i'
          ),
          total_record_count,
      )

    if unlabeled_record_count == total_record_count:
      raise ValueError(
          'There were no labels found in the dataset. Check the'
          'value passed in for the label value and label column'
          'are correct if your datasets has labels.'
      )

    if (
        self.runner_parameters.labeling_and_model_training_batch_size
        is not None
    ):
      if self.runner_parameters.labeling_and_model_training_batch_size < 1:
        raise ValueError(
            'labeling_and_model_training_batch_size must be set to '
            'a value greater than 0.'
        )

      if (
          self.runner_parameters.labeling_and_model_training_batch_size
          > total_record_count
      ):
        raise ValueError(
            'labeling_and_model_training_batch_size can not be greater than '
            f'the total number of examples. There are {total_record_count} '
            'examples and batch size is set to '
            f'{self.runner_parameters.labeling_and_model_training_batch_size}.'
        )

  def instantiate_and_fit_ensemble(
      self, unlabeled_record_count: int
  ) -> occ_ensemble.GmmEnsemble:
    """Creates and fits an OCC ensemble on the specified input data.

    Args:
      unlabeled_record_count: Number of unlabeled records in the table.

    Returns:
      A trained one class classifier ensemble.
    """

    ensemble_object = occ_ensemble.GmmEnsemble(
        ensemble_count=self.runner_parameters.ensemble_count,
        positive_threshold=self.runner_parameters.positive_threshold,
        negative_threshold=self.runner_parameters.negative_threshold,
    )

    records_per_occ = unlabeled_record_count // ensemble_object.ensemble_count
    batch_size = records_per_occ // self.runner_parameters.batches_per_model
    batch_size = np.min([batch_size, self.runner_parameters.max_occ_batch_size])

    logging.info('Batch size for OCC ensemble: %s', batch_size)

    unlabeled_data = self.data_loader.load_tf_dataset_from_bigquery(
        input_path=self.runner_parameters.input_bigquery_table_path,
        label_col_name=self.runner_parameters.label_col_name,
        where_statements=self.runner_parameters.where_statements,
        ignore_columns=self.runner_parameters.ignore_columns,
        batch_size=batch_size,
        label_column_filter_value=self.runner_parameters.unlabeled_data_value,
    )

    logging.info('Fitting ensemble.')
    ensemble_object.fit(
        train_x=unlabeled_data,
        batches_per_occ=self.runner_parameters.batches_per_model,
    )
    logging.info('Ensemble fit complete.')

    return ensemble_object

  def write_verbose_logs(
      self,
      features: np.ndarray,
      labels: np.ndarray,
      weights: np.ndarray,
  ) -> None:
    """Writes logs to Cloud Logging.

    Args:
      features: Set of all labeled features (after pseudo labels added).
      labels: Set of all positive and negative labels.
      weights: Weights corresponding to pseudo labels - this is the alpha
        parameter.
    """
    updated_label_counts = pd.DataFrame(labels).value_counts()
    logging.info('Updated label counts %s', updated_label_counts)

    logging.info('Test features shape: %s',
                 self.test_x.shape if self.test_x is not None else None)
    logging.info('Test labels shape: %s',
                 self.test_y.shape if self.test_y is not None else None)

    logging.info('Updated features shape: %s', features.shape)
    logging.info('Updated labels shape: %s', labels.shape)
    logging.info('Weights shape: %s', weights.shape)

    logging.info('Features sample: %s', features[:1])
    logging.info('Labels sample: %s', labels[:1])
    logging.info('Weights sample: %s', weights[:1])

  def evaluate_model(
      self,
      batch_number: Optional[int] = 0,
  ) -> None:
    """Evaluates the supervised model and writes logs to Cloud Logging.

    Args:
      batch_number: Integer corresponding to the batch of pseudo labeling and
        supervised model training the loop is in. We use this for calculating
        running averages for various metrics.
    """

    if self.test_x is None:
      raise ValueError(
          'There is no test set to evaluate on, double check '
          'the testing table or train/test split fraction.'
      )

    if self.supervised_model_object is None:
      raise ValueError(
          'Evaluate called without a trained supervised model. Ensure that '
          'run() has been called and the algorithm is not being ran in upload '
          'only mode.'
      )

    else:
      eval_results = self.supervised_model_object.supervised_model.evaluate(
          x=self.test_x, y=self.test_y, return_dict=True
      )
      if batch_number == 0:
        self.supervised_model_metrics = eval_results
        logging.info(
            'Supervised model evaluation %s',
            eval_results,
        )
      else:
        for metric_name, metric_value in eval_results.items():
          if self.supervised_model_metrics is not None:
            self.supervised_model_metrics[metric_name] += (
                metric_value - self.supervised_model_metrics[metric_name]
            ) / (batch_number + 1)

        logging.info(
            'Supervised model evaluation for current batch %s',
            eval_results,
        )

        logging.info(
            'Supervised model average evaluation: %s',
            self.supervised_model_metrics,
        )

  def _check_runner_parameters(self) -> None:
    """Checks parameter that are not related to the data tables.

    When developing workflows here, set parameters and warn users as
    needed. Throw errors when an essential parameter is missing or out of a
    required range.
    """
    if self.runner_parameters.test_dataset_holdout_fraction > 1:
      raise ValueError(
          'Can not use more than 100% of the data for the test '
          'set:'
          f' test_dataset_holdout_fraction={self.runner_parameters.test_dataset_holdout_fraction}'
      )

    if self.runner_parameters.test_dataset_holdout_fraction > 0.5:
      logging.warning(
          (
              'Using more than 50%% of the data for a test set, this can lead'
              ' to negative performance implications.'
              ' test_dataset_holdout_fraction = %s'
          ),
          self.runner_parameters.test_dataset_holdout_fraction,
      )
    if (
        self.runner_parameters.test_dataset_holdout_fraction > 0
        and self.runner_parameters.test_bigquery_table_path
    ):
      logging.warning(
          'Only a test holdout fraction and a single input table '
          'or an input and test table may be specified. Using the '
          'test table instead of the specified holdout fraction.'
      )
      self.runner_parameters.test_dataset_holdout_fraction = 0

    if self.runner_parameters.upload_only and (
        self.runner_parameters.test_dataset_holdout_fraction
        or self.runner_parameters.test_bigquery_table_path
    ):
      logging.warning(
          'A test set is not needed in upload only mode, '
          'test_dataset_holdout_fraction and test_bigquery_table_path will be '
          'ignored.'
      )
      self.runner_parameters.test_dataset_holdout_fraction = 0
      self.runner_parameters.test_bigquery_table_path = ''
      self.runner_parameters.output_gcs_uri = ''

    if (
        self.runner_parameters.upload_only
        and not self.runner_parameters.output_bigquery_table_path
    ):
      raise ValueError(
          'output_bigquery_table_path needs to be specified in '
          'upload_only mode.'
      )

  def preprocess_train_test_split(
      self,
      features: np.ndarray,
      labels: np.ndarray,
  ) -> Tuple[np.ndarray, np.ndarray]:
    """Creates the train and test sets depending on the train setting.

    Args:
      features: Testing features, numpy array.
      labels: Testing labels, numpy array.

    Returns:
      A sequence consisting of processed training and test sets in the order of
      train_x, train_y.
    """
    if len(features) != len(labels):
      raise ValueError(
          'Feature and label arrays have different lengths. '
          f'{len(features)} != {len(labels)}'
      )
    # Shuffle arrays, best practice before creating train/test splits.
    random_indices = np.random.permutation(len(features))
    features = features[random_indices]
    labels = labels[random_indices]

    if self.runner_parameters.train_setting == parameters.TrainSetting.PNU:
      ground_truth_label_indices = np.where(
          labels != self.runner_parameters.unlabeled_data_value
      )[0]
      label_count = int(
          len(ground_truth_label_indices)
          * self.runner_parameters.test_dataset_holdout_fraction
      )
      test_index_subset = np.random.choice(
          ground_truth_label_indices, size=label_count, replace=False
      )

      train_x = np.delete(features, test_index_subset, axis=0)
      train_y = np.delete(labels, test_index_subset, axis=0)

      test_x = features[test_index_subset]
      test_y = labels[test_index_subset]

    # TODO(b/247116870): Investigate the performance implications and user
    # interest for including some of the negative data in the training set
    # where we re-label as unlabeled data. This could improve the accuracy of
    # the end supervised classifier.
    elif self.runner_parameters.train_setting == parameters.TrainSetting.PU:
      # Uses all ground truth negative labels and the correct proportion of
      # positive labels for testing.
      positive_indices = np.where(
          labels == self.runner_parameters.positive_data_value
      )[0]
      negative_indices = np.where(
          labels == self.runner_parameters.negative_data_value
      )[0]

      positive_label_count = int(
          len(positive_indices)
          * self.runner_parameters.test_dataset_holdout_fraction
      )
      test_positive_indices = np.random.choice(
          positive_indices, size=positive_label_count, replace=False
      )

      train_x = np.delete(
          features,
          np.concatenate([test_positive_indices, negative_indices], axis=0),
          axis=0,
      )
      train_y = np.delete(
          labels,
          np.concatenate([test_positive_indices, negative_indices], axis=0),
          axis=0,
      )

      test_x = np.concatenate(
          [features[negative_indices], features[test_positive_indices]], axis=0
      )
      test_y = np.concatenate(
          [labels[negative_indices], labels[test_positive_indices]], axis=0
      )

    else:
      raise ValueError(
          'Unknown train setting for preparing train/test '
          f'datasets: {self.runner_parameters.train_setting}'
      )

    # TODO(b/247116870): Implement a dedicated function in the runner class
    # to load BQ test sets before evaluating the supervised model.
    if self.runner_parameters.test_bigquery_table_path:
      # Remove any unlabeled samples that may be in the test set.
      unlabeled_sample_filter = (
          f'{self.runner_parameters.test_label_col_name} != '
          f'{self.runner_parameters.unlabeled_data_value}'
      )
      if self.runner_parameters.where_statements is not None:
        unlabeled_sample_where_statements = (
            self.runner_parameters.where_statements + [unlabeled_sample_filter]
        )
      else:
        unlabeled_sample_where_statements = [unlabeled_sample_filter]
      test_dataset_size = self.data_loader.get_query_record_result_length(
          input_path=self.runner_parameters.test_bigquery_table_path,
          where_statements=unlabeled_sample_where_statements,
      )
      test_tf_dataset = self.data_loader.load_tf_dataset_from_bigquery(
          input_path=self.runner_parameters.test_bigquery_table_path,
          label_col_name=self.runner_parameters.test_label_col_name,
          where_statements=unlabeled_sample_where_statements,
          ignore_columns=self.runner_parameters.ignore_columns,
          batch_size=test_dataset_size,
      )

      test_x, test_y = test_tf_dataset.as_numpy_iterator().next()
      self.test_x = np.array(test_x)
      self.test_y = np.array(test_y)

      if not (
          np.any(test_y == self.runner_parameters.positive_data_value)
          and np.any(test_y == self.runner_parameters.negative_data_value)
      ):
        raise ValueError(
            'Positive and negative labels must be in the testing set. Please '
            'check the test table provided in the test_bigquery_table_path '
            'parameter.'
        )
    else:
      if self.test_x is not None:
        self.test_x = np.concatenate([self.test_x, test_x], axis=0)
        self.test_y = np.concatenate([self.test_y, test_y], axis=0)
      else:
        self.test_x = test_x
        self.test_y = test_y

    # Adjust the testing labels to values of 1 and 0 to align with the class
    # the supervised model is trained on.
    self.test_y[self.test_y == self.runner_parameters.positive_data_value] = 1
    self.test_y[self.test_y == self.runner_parameters.negative_data_value] = 0

    return (train_x, train_y)

  def train_supervised_model(
      self,
      *,
      features: np.ndarray,
      labels: np.ndarray,
      weights: np.ndarray,
  ) -> None:
    """Trains a supervised model.

    This function can be called in a batch manner if the supervised model
    supports batch updates.

    Args:
      features: Pseudo labeled and ground truth features.
      labels: Pseudo and ground truth labels.
      weights: Weights for pseudo labeled data.

    Raises:
      ValueError if there is the supervised model object is None.
    """
    if self.supervised_model_object is None:
      raise ValueError('No supervised model to train.')
    logging.info('Supervised model training started.')
    self.supervised_model_object.train(
        features=features, labels=labels, weights=weights
    )
    logging.info('Supervised model training completed.')

  def run(self) -> None:
    """Runs the anomaly detection algorithm on a BigQuery table."""
    logging.info('SPADE training started.')

    self._check_runner_parameters()

    total_record_count = self.data_loader.get_query_record_result_length(
        input_path=self.runner_parameters.input_bigquery_table_path,
        where_statements=self.runner_parameters.where_statements,
    )

    unlabeled_record_count_filter = (
        f'{self.runner_parameters.label_col_name} = '
        f'{self.runner_parameters.unlabeled_data_value}'
    )
    if self.runner_parameters.where_statements:
      unlabeled_record_count_where_statements = [
          self.runner_parameters.where_statements
      ] + [unlabeled_record_count_filter]
    else:
      unlabeled_record_count_where_statements = [unlabeled_record_count_filter]

    unlabeled_record_count = self.data_loader.get_query_record_result_length(
        input_path=self.runner_parameters.input_bigquery_table_path,
        where_statements=unlabeled_record_count_where_statements,
    )

    self.check_data_tables(
        total_record_count=total_record_count,
        unlabeled_record_count=unlabeled_record_count,
    )

    ensemble_object = self.instantiate_and_fit_ensemble(
        unlabeled_record_count=unlabeled_record_count
    )

    batch_size = (
        self.runner_parameters.labeling_and_model_training_batch_size
        or total_record_count
    )
    tf_dataset = self.data_loader.load_tf_dataset_from_bigquery(
        input_path=self.runner_parameters.input_bigquery_table_path,
        label_col_name=self.runner_parameters.label_col_name,
        where_statements=self.runner_parameters.where_statements,
        ignore_columns=self.runner_parameters.ignore_columns,
        batch_size=batch_size,
    )
    tf_dataset = tf_dataset.as_numpy_iterator()

    for batch_number, (features, labels) in enumerate(tf_dataset):
      logging.info(
          'Labeling and supervised model training batch number: %s',
          batch_number,
      )
      logging.info(
          'Batch size: %s',
          len(features),
      )

      train_x, train_y = self.preprocess_train_test_split(
          features=features,
          labels=labels,
      )

      logging.info('Labeling started.')
      updated_features, updated_labels, weights = ensemble_object.pseudo_label(
          features=train_x,
          labels=train_y,
          positive_data_value=self.runner_parameters.positive_data_value,
          negative_data_value=self.runner_parameters.negative_data_value,
          unlabeled_data_value=self.runner_parameters.unlabeled_data_value,
          alpha=self.runner_parameters.alpha,
          verbose=self.runner_parameters.verbose,
      )
      logging.info('Labeling completed.')

      # Upload batch of pseudo labels, will append when called more than once.
      if self.runner_parameters.output_bigquery_table_path:
        self.data_loader.upload_dataframe_as_bigquery_table(
            features=updated_features, labels=updated_labels
        )

      # TODO(b/247116870): Create two logging functions, one for batch and one
      # for the end of SPADE training (reporting job level metrics such as AUC).
      if self.runner_parameters.verbose:
        self.write_verbose_logs(
            features=updated_features,
            labels=updated_labels,
            weights=weights,
        )

      if not self.runner_parameters.upload_only:
        self.train_supervised_model(
            features=updated_features,
            labels=updated_labels,
            weights=weights,
        )

    if not self.runner_parameters.upload_only:
      self.evaluate_model()
      if self.supervised_model_object is not None:
        self.supervised_model_object.save(
            save_location=self.runner_parameters.output_gcs_uri
        )

    logging.info('SPADE training completed.')
